# Copyright (c) 2026, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Tuple

import numpy as np
from megatron.core.datasets.gpt_dataset import GPTDataset
from megatron.core.datasets.indexed_dataset import IndexedDataset
from megatron.core.datasets.utils import Split

from megatron.bridge.training.config import GPTFIMDatasetConfig


logger = logging.getLogger(__name__)


class GPTFIMDataset(GPTDataset):
    """FIM (Fill In The Middle) GPT Dataset

    Args:
        indexed_dataset (IndexedDataset): The IndexedDataset around which to build the
        MegatronDataset

        indexed_indices (np.ndarray): The set of the documents indices to expose

        num_samples (int): The number of samples to draw from the indexed dataset

        index_split (Split): The indexed_indices Split

        config (GPTFIMDatasetConfig): The GPT-specific container for all config sourced parameters
    """

    def __init__(
        self,
        indexed_dataset: IndexedDataset,
        dataset_path: str,
        indexed_indices: np.ndarray,
        num_samples: int,
        index_split: Split,
        config: GPTFIMDatasetConfig,
    ) -> None:
        super().__init__(indexed_dataset, dataset_path, indexed_indices, num_samples, index_split, config)

        self.np_rng = np.random.RandomState(seed=self.config.random_seed)
        logger.info(f"Initialized FIM RNG with seed = {self.config.random_seed}")
        # get FIM params
        self.fim_rate = self.config.fim_rate
        self.fim_spm_rate = self.config.fim_spm_rate
        self.fragment_fim_rate = self.config.fim_fragment_rate
        fim_split_sample = self.config.fim_split_sample
        self.no_fim_prefix = self.config.fim_no_prefix
        if fim_split_sample:
            fim_split_sample_ids = self.config.tokenizer._tokenizer.tokens_to_ids(fim_split_sample)
            assert isinstance(fim_split_sample_ids, int) or len(fim_split_sample_ids) == 1
            self.fim_split_sample = (
                fim_split_sample_ids if isinstance(fim_split_sample_ids, int) else fim_split_sample_ids[0]
            )
        else:
            self.fim_split_sample = None

        # get extra tokens ids
        fim_tokens = self.config.fim_extra_tokens
        fim_tokens = [
            fim_tokens["prefix"],
            fim_tokens["middle"],
            fim_tokens["suffix"],
            fim_tokens["pad"],
            fim_tokens["eod"],
        ]
        fim_tokens_ids = self.config.tokenizer._tokenizer.tokens_to_ids(fim_tokens)
        (
            self.prefix_tok_id,
            self.middle_tok_id,
            self.suffix_tok_id,
            self.pad_tok_id,
            self.eod_tok_id,
        ) = fim_tokens_ids

    def _query_document_sample_shuffle_indices(self, idx: int) -> Tuple[np.ndarray, np.ndarray]:
        """Get the text (token ids) and document ids for a given index

        Args:
            idx (int): The index into the dataset

        Returns:
            Tuple[np.ndarray, np.ndarray]: The text ids and document ids
        """
        # Do the shuffle mapping
        idx = self.shuffle_index[idx]

        # Get the beginning and end documents and offsets
        doc_index_beg, doc_index_beg_offset = self.sample_index[idx]
        doc_index_end, doc_index_end_offset = self.sample_index[idx + 1]

        document_ids = []
        sample_parts = []

        # Sample spans a single document
        if doc_index_beg == doc_index_end:
            # Add the document id
            document_ids.append(self.document_index[doc_index_beg])

            # Add the entire sample
            sample_parts.append(
                self.dataset.get(
                    self.document_index[doc_index_beg],
                    offset=doc_index_beg_offset,
                    length=doc_index_end_offset - doc_index_beg_offset + 1,
                )
            )

        # Sample spans multiple documents
        else:
            for i in range(doc_index_beg, doc_index_end + 1):
                # Add the document id
                document_ids.append(self.document_index[i])

                # Add the sample part
                offset = 0 if i > doc_index_beg else doc_index_beg_offset
                length = None if i < doc_index_end else doc_index_end_offset + 1
                sample_parts.append(self.dataset.get(self.document_index[i], offset=offset, length=length))

        sample = np.concatenate(sample_parts)

        sample_len = sample.shape[0]
        segment_breaks = np.argwhere(sample == self.eod_tok_id)

        if segment_breaks.shape != (0, 1):  # then there is an EOD token in this example
            curr_start_position = 0
            new_samples = []
            for loc in np.nditer(segment_breaks):
                # Only permute non-empty segments.
                if loc - curr_start_position > 0:
                    # permute {prefix, suffix, middle} or {suffix, prefix, middle}
                    permuted = self._fim_split_and_permute_sequence(sample[curr_start_position:loc])
                    new_samples += [permuted, [self.eod_tok_id]]

                curr_start_position = loc + 1  # jump over the EOD token
            # Permute the segment after the last EOD
            permuted = self._fim_split_and_permute_sequence(sample[curr_start_position:])
            new_samples.append(permuted)

            sample = np.concatenate(new_samples)
        else:
            sample = self._fim_split_and_permute_sequence(sample)

        diff = sample.shape[0] - sample_len
        if diff > 0:  # too long
            sample = sample[:sample_len]
        elif diff < 0:  # too short
            sample = np.concatenate([sample, np.full((-1 * diff), self.pad_tok_id)])

        assert sample.shape[0] == sample_len

        return (np.array(sample, dtype=np.int64), np.array(document_ids, dtype=np.int64))

    def _fim_permute_sequence(self, sequence, rate):
        return self._permute(
            sequence,
            rate,
            self.fim_spm_rate,
            self.config.tokenizer,
            truncate_or_pad=False,
            suffix_tok_id=self.suffix_tok_id,
            prefix_tok_id=self.prefix_tok_id,
            middle_tok_id=self.middle_tok_id,
            pad_tok_id=self.pad_tok_id,
            no_fim_prefix=self.no_fim_prefix,
        )

    def _fim_split_and_permute_sequence(self, sequence):
        """
        If self.fim_split_sample is not None, split the sequence.
        Then apply FIM on the fragments, or the whole sequence if self.fim_split_sample is None.
        """
        if self.fim_split_sample is None:
            return self._fim_permute_sequence(sequence, self.fim_rate)
        # fim_split_sample is set: split the sample on this token and permute each fragment separately.
        # Typically, if each sample is a repository, then we split again on the file level.
        # Each fragment is a file, and we permute the files.
        fragment_breaks = np.argwhere(sequence == self.fim_split_sample)
        if fragment_breaks.shape == (0, 1):
            # no split token in this sample
            return self._fim_permute_sequence(sequence, self.fim_rate)
        if not self.np_rng.binomial(1, self.fim_rate):
            # don't do FIM preproc
            return sequence
        # Do FIM on each fragment
        curr_start_position = 0
        new_samples = []
        for loc in np.nditer(fragment_breaks):
            if loc - curr_start_position > 0:
                permuted = self._fim_permute_sequence(sequence[curr_start_position:loc], self.fragment_fim_rate)
                new_samples += [permuted, [self.fim_split_sample]]
            curr_start_position = loc + 1  # Jump over the split token
        # Permute the segment after the last split token
        permuted = self._fim_permute_sequence(sequence[curr_start_position:], self.fragment_fim_rate)
        new_samples.append(permuted)

        return np.concatenate(new_samples)

    def _permute(
        self,
        sample,
        fim_rate,
        fim_spm_rate,
        tokenizer,
        truncate_or_pad=True,
        suffix_tok_id=None,
        prefix_tok_id=None,
        middle_tok_id=None,
        pad_tok_id=None,
        no_fim_prefix=None,
    ):
        """
        Take in a sample (np array w/ size (0,chunklength)) and perform a FIM transformation on it.
        Maintain the same sample length (if transform creates a few extra tokens, drop them).
        """
        if self.np_rng.binomial(1, fim_rate):  # sample bernoulli dist
            contents = tokenizer._tokenizer.ids_to_text(sample)

            # Do not apply FIM if the sample starts with no_fim_prefix
            if no_fim_prefix is not None and contents.startswith(no_fim_prefix):
                return sample

            try:
                # A boundary can be =0 (prefix will be empty)
                # a boundary can be =len(contents) (suffix will be empty)
                # The two boundaries can be equal (middle will be empty)
                boundaries = list(self.np_rng.randint(low=0, high=len(contents) + 1, size=2))
                boundaries.sort()
            except ValueError as e:
                print(len(contents), contents)
                print(e)
                raise e

            prefix = contents[: boundaries[0]]
            middle = contents[boundaries[0] : boundaries[1]]
            suffix = contents[boundaries[1] :]

            prefix = np.array([*tokenizer._tokenizer.text_to_ids(prefix)], dtype=np.int64)
            middle = np.array([*tokenizer._tokenizer.text_to_ids(middle)], dtype=np.int64)
            suffix = np.array([*tokenizer._tokenizer.text_to_ids(suffix)], dtype=np.int64)

            # here we truncate each given segment to fit the same length as it was before
            # A consequence is that we never reach the end of a file?
            # we should rather truncate at the context-level
            if truncate_or_pad:
                # need to make same length as the input. Take the 3 sentinel tokens into account
                new_length = suffix.shape[0] + prefix.shape[0] + middle.shape[0] + 3
                diff = new_length - sample.shape[0]
                if diff > 0:  # too long
                    if (
                        suffix.shape[0] <= diff
                    ):  # if there's no space to truncate the suffix: stop and report it. atm i should have stopped this from happening
                        return sample
                    suffix = suffix[: suffix.shape[0] - diff]
                elif diff < 0:  # too short
                    suffix = np.concatenate([suffix, np.full((-1 * diff), pad_tok_id)])

            if self.np_rng.binomial(1, fim_spm_rate):
                # SPM (variant 2 from FIM paper)
                new_sample = np.concatenate([[prefix_tok_id, suffix_tok_id], suffix, [middle_tok_id], prefix, middle])
            else:
                # PSM
                new_sample = np.concatenate(
                    [[prefix_tok_id], prefix, [suffix_tok_id], suffix, [middle_tok_id], middle]
                )

        else:
            # don't do FIM preproc
            new_sample = sample

        return new_sample



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import OrderedDict

import torch
from megatron.core.inference.inference_request import InferenceRequest
from megatron.core.inference.text_generation_controllers.simple_text_generation_controller import (
    SimpleTextGenerationController,
)


class TokenizerWrapper:
    """Wrapper around tokenizer to provide a unified interface for inference."""

    # pylint: disable=C0115,C0116
    def __init__(self, tokenizer):
        self.eod = tokenizer.eos_token_id
        self.vocab_size = None
        self._tokenizer = tokenizer

    def detokenize(self, tokens):
        # pylint: disable=C0115,C0116
        return self._tokenizer.decode(tokens, skip_special_tokens=False)

    def tokenize(self, prompt):
        # pylint: disable=C0115,C0116
        return self._tokenizer.encode(prompt, add_special_tokens=False)


class VLMTextGenerationController(SimpleTextGenerationController):
    """Text generation controller for VLM models with image processing support."""

    # pylint: disable=C0115,C0116
    def __init__(self, inference_wrapped_model, tokenizer, image_processor):
        super().__init__(inference_wrapped_model, TokenizerWrapper(tokenizer))
        self.image_processor = image_processor

    def tokenize_prompt(self, prompt: str, image):
        # pylint: disable=C0115,C0116
        tokens = self.tokenizer.tokenize(prompt)
        if image is None:
            image_dict = dict(
                pixel_values=torch.zeros(
                    1, 4, 3, self.image_processor.size["height"], self.image_processor.size["width"]
                ),
                aspect_ratio_ids=torch.tensor([0], dtype=torch.long),
                num_tiles=[0],
            )
        else:
            image_dict = self.image_processor.preprocess(image, return_tensors="pt")
            image_dict = {
                k: v[0] for k, v in image_dict.items() if k in ["pixel_values", "aspect_ratio_ids", "num_tiles"]
            }
        return tokens, image_dict

    def prep_inference_input(
        self,
        prompts_tokens: torch.Tensor,
        active_requests: OrderedDict[str, InferenceRequest],
        use_attention_mask: bool = False,
    ):
        """Preparing input data for inference, using respective wrapper's prep_inference_input method

        Args:
            prompts_tokens (torch.Tensor): A tensor of shape [batch_size, max_sequence_length]
            active_requests (OrderedDict[int, InferenceRequest]): The input active requests
            use_attention_mask (bool): Whether to use an attention mask. Should be set to True only
                when exclusively doing prefill (no decode) with variable prompt lengths.
                Currently unused and added to match an expected interface in Mcore.
        """
        images = list(map(lambda request: request.encoder_prompt, active_requests.values()))

        return self.inference_wrapped_model.prep_inference_input(
            prompts_tokens=prompts_tokens,
            image_dict=images,
        )


class QwenVLTextGenerationController(VLMTextGenerationController):
    """Text generation controller for QwenVL model"""

    def __init__(self, inference_wrapped_model, tokenizer, image_processor, processor):
        super().__init__(inference_wrapped_model, tokenizer, image_processor)

        class QwenVLTokenizer(TokenizerWrapper):
            # pylint: disable=C0115,C0116
            def detokenize(self, tokens):
                new_tokens = []
                for token in tokens:
                    if token == 151652:
                        new_tokens.append(token)
                        new_tokens.append(151655)
                    elif token != 151655:
                        new_tokens.append(token)
                return self._tokenizer.decode(new_tokens, skip_special_tokens=False)

        self.tokenizer = QwenVLTokenizer(tokenizer)
        self.processor = processor

    def tokenize_prompt(self, prompt: str, image):
        """Tokenize prompt and process image following the VLM controller API"""
        inputs = self.processor(
            text=[prompt],
            images=image,
            padding=True,
            return_tensors="pt",
        )

        tokens = inputs["input_ids"][0]

        if "pixel_values" in inputs:
            image_dict = {
                "pixel_values": inputs["pixel_values"],
                "image_grid_thw": inputs["image_grid_thw"],
            }
        else:
            image_dict = None
        return tokens.tolist(), image_dict



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from dataclasses import dataclass, field
from typing import List, Literal, Optional

from torch import nn

from megatron.bridge.peft.base import PEFT
from megatron.bridge.peft.dora_layers import DoRALinear, ParallelLinearDoRAAdapter
from megatron.bridge.peft.module_matcher import ModuleMatcher
from megatron.bridge.peft.utils import get_adapter_attributes_from_linear


logger = logging.getLogger(__name__)


@dataclass
class DoRA(PEFT, ModuleMatcher):
    """
    Implements the DoRA (Weight-Decomposed LowRank Adaptation) module for parameter-efficient fine-tuning.

    DoRA decomposes pre-trained weight into magnitude and direction, and uses a low-rank projection in the
    directional component to adapt the weights of a pre-trained model to a new downstream task.
    This class facilitates the application of DoRA to specific modules within the model architecture.

    Args:
        target_modules (List[str], optional): A list of module names to apply DoRA to.
            Defaults to all linear layers ['linear_qkv', 'linear_proj', 'linear_fc1', 'linear_fc2'].
                - 'linear_qkv': Apply DoRA to the fused linear layer used for query, key, and value projections
                                in self-attention.
                - 'linear_proj': Apply DoRA to the linear layer used for projecting the output of self-attention.
                - 'linear_fc1': Apply DoRA to the first fully-connected layer in MLP.
                - 'linear_fc2': Apply DoRA to the second fully-connected layer in MLP.
            Target modules can also contain wildcards. For example, you can specify
                target_modules=['*.layers.0.*.linear_qkv', '*.layers.1.*.linear_qkv'] to add DoRA to only linear_qkv
                on the first two layers.
        dim (int): Dimension of the low-rank projection space. Defaults to 32.
        alpha (int): Weighting factor for the low-rank projection. Defaults to 64.
        dropout (float): Dropout rate for the low-rank projection. Defaults to 0.0.
        dropout_position (Literal['pre', 'post'], optional): Position for applying dropout.
            Can be 'pre' (before the low-rank projection) or 'post' (after). Defaults to 'pre'.
        lora_A_init_method (str): Initialization method for the low-rank matrix A. Defaults to "xavier".
        lora_B_init_method (str): Initialization method for the low-rank matrix B. Defaults to "zero".
    """

    target_modules: List[str] = field(
        default_factory=lambda: ["linear_qkv", "linear_proj", "linear_fc1", "linear_fc2"]
    )
    dim: int = 32
    alpha: int = 64
    dropout: float = 0.0
    dropout_position: Literal["pre", "post"] = "pre"
    lora_A_init_method: str = "xavier"
    lora_B_init_method: str = "zero"

    def __post_init__(self):
        """Initialize attributes from parent classes and validate configuration."""
        assert self.dropout_position == "pre", (
            "DoRA only supports pre-adapter dropout at this time. Please set DoRA(..., dropout_position='pre')"
        )

    def transform(self, m: nn.Module, name: Optional[str] = None, prefix: Optional[str] = None) -> nn.Module:
        """
        Applies DoRA to a specific module within the model architecture.

        Args:
            m (nn.Module): The module to apply DoRA to.
            name (str, optional): Name of the module (if applicable). Defaults to None.
            prefix (str, optional): Prefix for the module name (if applicable). Defaults to None.

        Returns:
            nn.Module: The modified module with DoRA applied, or the original module if not a target.
        """
        # Skip already transformed modules
        if isinstance(m, DoRALinear):
            return m

        if (ans := self.match(m, name, prefix)) is not None:
            (match, full_name) = ans
            attrs = get_adapter_attributes_from_linear(m)
            logger.info(f"Adding DoRA to: {full_name}")
            adapter = ParallelLinearDoRAAdapter(
                attrs.in_features,
                attrs.out_features,
                self.dim,
                base_linear_name=full_name,
                activation="identity",
                norm_type=None,
                column_init_method=self.lora_A_init_method,
                row_init_method=self.lora_B_init_method,
                gather_output=False,
                input_is_parallel=attrs.input_is_parallel,
                dropout=self.dropout,
                dropout_position=self.dropout_position,
                model_parallel_config=getattr(m, "config", None),
                alpha=self.alpha,
                disable_tensor_parallel_comm=attrs.disable_tensor_parallel_comm,
                disable_sequence_parallel_comm=attrs.disable_sequence_parallel_comm,
                base_linear_is_parallel=attrs.base_linear_is_parallel,
            )
            return DoRALinear(m, adapter)
        return m



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from dataclasses import dataclass, fields
from typing import TYPE_CHECKING, Any, Optional

import modelopt.torch.distill as mtd
import modelopt.torch.distill.plugins.megatron as mtd_mcore
from megatron.core.models.gpt import GPTModel as MCoreGPTModel

from megatron.bridge.models.gpt_provider import GPTModelProvider
from megatron.bridge.models.mamba.mamba_provider import MambaModelProvider
from megatron.bridge.models.transformer_config import TransformerConfig


if TYPE_CHECKING:
    from megatron.bridge.training.post_training.distillation import ModelOptDistillConfig


logger = logging.getLogger(__name__)


@dataclass
class DistillationProvider(TransformerConfig):
    """Provider for Megatron Core GPT models in distillation mode.

    Please use `convert_to_distillation_provider()` to create an instance of this class.
    """

    teacher: Optional[GPTModelProvider | MambaModelProvider] = None
    kd_config: Optional["ModelOptDistillConfig"] = None

    def __init__(self, *args, **kwargs):
        raise NotImplementedError("Use `convert_to_distillation_provider()` to create an instance of this class.")

    def __post_init__(self):
        assert getattr(self, "teacher", None) is not None, "Teacher model must be provided."

        shared_attrs = [
            "tensor_model_parallel_size",
            "pipeline_model_parallel_size",
            "context_parallel_size",
            "seq_length",
            "pipeline_dtype",
        ]
        for attr in shared_attrs:
            if getattr(self, attr) != getattr(self.teacher, attr):
                raise ValueError(f"Student and teacher providers must have the same {attr}.")

        # Logits are overwritten in-place when TE cross-entropy loss is enabled, so switch it back to native version.
        self.cross_entropy_fusion_impl = "native"

        # Hack to dynamically subclass other providers and still use their methods
        self._super_class = self.__class__.__bases__[0]

    def provide(self, pre_process=None, post_process=None, vp_stage=None) -> MCoreGPTModel:
        """Configure and instantiate a ModelOpt DistillationModel based on this configuration.

        Args:
            pre_process: Whether to include pre-processing in the model, defaults to first pipeline stage
            post_process: Whether to include post-processing in the model, defaults to last pipeline stage
            vp_stage: Virtual pipeline stage

        Returns:
            MCoreGPTModel: Configured ModelOpt DistillationModel instance
        """
        if vp_stage is not None:
            raise ValueError("ModelOpt KD currently does not support virtual-pipeline parallel.")

        student_model = self._super_class.provide(self, pre_process, post_process, vp_stage)
        # Hack to get teacher's pre-wrap hooks called to potentially load HF weights
        teacher_model = self.teacher.provide_distributed_model(wrap_with_ddp=False, mixed_precision_wrapper=None)[0]

        kd_cfg = mtd_mcore.setup_distillation_config(self.kd_config, student_model.config, teacher_model.config)
        modelopt_cfg = {
            "teacher_model": teacher_model,
            "criterion": kd_cfg.criterion,
            "loss_balancer": kd_cfg.loss_balancer,
        }
        kd_model = mtd.convert(student_model, mode=[("kd_loss", modelopt_cfg)])
        mtd_mcore.adjust_distillation_model_for_mcore(kd_model, kd_cfg)

        return kd_model

    def to_cfg_dict(self) -> dict[str, Any]:
        """Custom method to save equivalent to the original provider class.

        Used by `_ConfigContainerBase` to serialize the main `ConfigContainer` to YAML.
        There is no need to restore a `DistillationProvider` from the run config file, as
        it can always be re-converted using the original student provider.

        Returns:
            Dictionary representation of this provider class
        """
        from megatron.bridge.training.utils.config_utils import _ConfigContainerBase

        result = {"_target_": f"{self._super_class.__module__}.{self._super_class.__qualname__}"}
        # Use fields from the actual student provider class, not DistillationProvider.
        # DistillationProvider's __dataclass_fields__ only includes TransformerConfig fields
        # (set at class definition time), missing GPTModelProvider-level fields like
        # vocab_size, share_embeddings_and_output_weights, etc.
        for field in fields(self._super_class):
            if field.name.startswith("_"):
                continue
            result[field.name] = _ConfigContainerBase._convert_value_to_dict(getattr(self, field.name))
        return result

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        # Mirror to teacher if it has that attribute
        if hasattr(self.teacher, name):
            setattr(self.teacher, name, value)


def convert_to_distillation_provider(
    student_provider: GPTModelProvider | MambaModelProvider,
    teacher_provider: GPTModelProvider | MambaModelProvider,
    kd_config: Optional["ModelOptDistillConfig"] = None,
) -> "DistillationProvider":
    """Convert a given model provider to a DistillationProvider."""

    assert isinstance(student_provider, (GPTModelProvider, MambaModelProvider)), (
        "Student provider must be a subclass of GPTModelProvider or MambaModelProvider."
    )
    assert isinstance(teacher_provider, (GPTModelProvider, MambaModelProvider)), (
        "Teacher provider must be a subclass of GPTModelProvider or MambaModelProvider."
    )

    DistillationProvider.__bases__ = (type(student_provider),)
    student_provider.__class__ = DistillationProvider

    student_provider.teacher = teacher_provider
    student_provider.kd_config = kd_config
    student_provider.__post_init__()

    return student_provider



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import torch
from megatron.core.models.gpt.gpt_model import GPTModel
from transformers import MistralForCausalLM

from megatron.bridge.models.conversion.mapping_registry import MegatronMappingRegistry
from megatron.bridge.models.conversion.model_bridge import MegatronModelBridge
from megatron.bridge.models.conversion.param_mapping import (
    AutoMapping,
    GatedMLPMapping,
    QKVMapping,
)
from megatron.bridge.models.hf_pretrained.causal_lm import PreTrainedCausalLM
from megatron.bridge.models.mistral.mistral_provider import MistralModelProvider


@MegatronModelBridge.register_bridge(source=MistralForCausalLM, target=GPTModel)
class MistralBridge(MegatronModelBridge):
    """
    Megatron Bridge for Mistral Causal LM.

    As a user you would not use this bridge directly, but through `AutoBridge`.

    Example:
        >>> from megatron.bridge import AutoBridge
        >>> bridge = AutoBridge.from_hf_pretrained("mistralai/Mistral-7B-Instruct")
        >>> provider = bridge.to_megatron_provider()
    """

    def provider_bridge(self, hf_pretrained: PreTrainedCausalLM) -> MistralModelProvider:
        hf_config = hf_pretrained.config

        if getattr(hf_config, "rope_scaling", None) is not None and hf_config.rope_scaling.get("rope_type") == "yarn":
            # Apply Mistral customize rope scaling
            cls = partial(MistralModelProvider, scale_factor=hf_config.rope_scaling.get("factor", 8.0))
        else:
            cls = MistralModelProvider

        window_size, cp_comm_type = (None, None)
        if getattr(hf_config, "sliding_window", None) is not None:
            window_size = [hf_config.sliding_window, 0]
            cp_comm_type = "a2a"

        provider = cls(
            num_layers=hf_config.num_hidden_layers,
            hidden_size=hf_config.hidden_size,
            ffn_hidden_size=hf_config.intermediate_size,
            num_attention_heads=hf_config.num_attention_heads,
            init_method_std=hf_config.initializer_range,
            layernorm_epsilon=hf_config.rms_norm_eps,
            num_query_groups=hf_config.num_key_value_heads,
            seq_length=hf_config.max_position_embeddings,
            rotary_base=hf_config.rope_theta,
            gated_linear_unit=True,
            make_vocab_size_divisible_by=self.make_vocab_size_divisible_by(hf_config.vocab_size),
            window_size=window_size,
            cp_comm_type=cp_comm_type,
            share_embeddings_and_output_weights=getattr(hf_config, "tie_word_embeddings", False),
            fp16=(self.dtype_from_hf(hf_config, default=torch.float32) == torch.float16),
            bf16=(self.dtype_from_hf(hf_config, default=torch.float32) == torch.bfloat16),
            params_dtype=self.dtype_from_hf(hf_config, default=torch.float32),
            vocab_size=hf_config.vocab_size,
            kv_channels=getattr(hf_config, "head_dim", None),
        )

        return provider

    def mapping_registry(self) -> MegatronMappingRegistry:
        # Return MegatronMappingRegistry containing parameter mappings from Megatron to HF format
        # First create simple 1:1 parameter mappings using a dictionary for readability

        # Dictionary maps Megatron parameter names -> HF parameter names
        # Supports wildcard (*) patterns for layer-specific parameters
        param_mappings = {
            "embedding.word_embeddings.weight": "model.embed_tokens.weight",
            "output_layer.weight": "lm_head.weight",
            "decoder.final_layernorm.weight": "model.norm.weight",
            "decoder.layers.*.self_attention.linear_qkv.layer_norm_weight": "model.layers.*.input_layernorm.weight",
            "decoder.layers.*.mlp.linear_fc1.layer_norm_weight": "model.layers.*.post_attention_layernorm.weight",
            "decoder.layers.*.self_attention.linear_proj.weight": "model.layers.*.self_attn.o_proj.weight",
            "decoder.layers.*.mlp.linear_fc2.weight": "model.layers.*.mlp.down_proj.weight",
        }

        mapping_list = []
        # Convert each dictionary entry to AutoMapping(megatron_param, hf_param)
        for megatron_param, hf_param in param_mappings.items():
            mapping_list.append(AutoMapping(megatron_param=megatron_param, hf_param=hf_param))

        # Add special mappings that require parameter concatenation/transformation
        mapping_list.extend(
            [
                # QKV: Combine separate Q, K, V matrices into single QKV matrix
                QKVMapping(
                    megatron_param="decoder.layers.*.self_attention.linear_qkv.weight",
                    q="model.layers.*.self_attn.q_proj.weight",
                    k="model.layers.*.self_attn.k_proj.weight",
                    v="model.layers.*.self_attn.v_proj.weight",
                ),
                # Gated MLP: Combine gate and up projection matrices into single FC1 matrix
                GatedMLPMapping(
                    megatron_param="decoder.layers.*.mlp.linear_fc1.weight",
                    gate="model.layers.*.mlp.gate_proj.weight",
                    up="model.layers.*.mlp.up_proj.weight",
                ),
            ]
        )

        return MegatronMappingRegistry(*mapping_list)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from megatron.core.models.gpt.gpt_model import GPTModel

from megatron.bridge.models.conversion.mapping_registry import MegatronMappingRegistry
from megatron.bridge.models.conversion.model_bridge import MegatronModelBridge
from megatron.bridge.models.conversion.param_mapping import (
    AutoMapping,
    GatedMLPMapping,
    QKVMapping,
)
from megatron.bridge.models.hf_pretrained.causal_lm import PreTrainedCausalLM
from megatron.bridge.models.llama.llama_provider import Llama31ModelProvider
from megatron.bridge.models.llama_nemotron.llama_nemotron_provider import LlamaNemotronHeterogeneousProvider


@MegatronModelBridge.register_bridge(source="DeciLMForCausalLM", target=GPTModel)
class LlamaNemotronBridge(MegatronModelBridge):
    """
    Megatron Bridge for Heterogeneous Llama-Nemotron models (Super/Ultra).

    This bridge handles heterogeneous Llama-Nemotron models that use the DeciLMForCausalLM
    architecture with block_configs for heterogeneous layer specifications. These models
    require special handling because:

    1. They use custom modeling code (DeciLMForCausalLM) loaded via auto_map
    2. They have heterogeneous block configurations (different layers have different specs)
    3. They require trust_remote_code=True to load from HuggingFace

    Supported models (examples):
    - nvidia/Llama-3_3-Nemotron-Super-49B-v1 (80 layers, 8192 hidden)
    - nvidia/Llama-3_3-Nemotron-Super-49B-v1_5 (updated v1.5 release)
    - nvidia/Llama-3_1-Nemotron-Ultra-253B-v1 (162 layers, 16384 hidden)

    Homogeneous Llama-Nemotron models (Nano/70B) use standard LlamaForCausalLM
    architecture and are handled by the regular LlamaBridge.

    Example:
        >>> from megatron.bridge import AutoBridge
        >>> # DeciLMForCausalLM models will automatically use this bridge
        >>> bridge = AutoBridge.from_hf_pretrained(
        ...     "nvidia/Llama-3_3-Nemotron-Super-49B-v1_5",
        ...     trust_remote_code=True
        ... )
        >>> provider = bridge.to_megatron_provider()
    """

    def provider_bridge(self, hf_pretrained: PreTrainedCausalLM) -> Llama31ModelProvider:
        hf_config = hf_pretrained.config
        # Validate heterogeneous DeciLM (NAS) config and select provider
        if not (hasattr(hf_config, "block_configs") and hf_config.block_configs):
            num_layers = getattr(hf_config, "num_hidden_layers", "unknown")
            raise ValueError(
                "LlamaNemotronBridge only handles heterogeneous models with block_configs. "
                f"Model with {num_layers} layers and no block_configs should use LlamaBridge."
            )
        archs = set(getattr(hf_config, "architectures", []) or [])
        auto_map = getattr(hf_config, "auto_map", {}) or {}
        is_decilm = ("DeciLMForCausalLM" in archs) or (
            auto_map.get("AutoModelForCausalLM", "").endswith("DeciLMForCausalLM")
        )
        if not is_decilm:
            raise ValueError("Unsupported heterogeneous architecture for LlamaNemotronBridge; expected DeciLM.")

        # Calculate num_query_groups for heterogeneous models
        # For heterogeneous models, GQA is defined in each block config
        # We assume block 0 has a non-no-op attention layer
        num_query_groups = hf_config.num_key_value_heads
        if hasattr(hf_config, "block_configs") and hf_config.block_configs:
            # Extract from block_configs[0].attention.n_heads_in_group
            block_0 = hf_config.block_configs[0]
            if hasattr(block_0, "attention") and hasattr(block_0.attention, "n_heads_in_group"):
                n_heads_in_group = block_0.attention.n_heads_in_group
                num_query_groups = hf_config.num_attention_heads // n_heads_in_group

        # Prepare kwargs for provider creation
        provider_kwargs = dict(
            num_layers=hf_config.num_hidden_layers,
            hidden_size=hf_config.hidden_size,
            ffn_hidden_size=hf_config.intermediate_size,
            num_attention_heads=hf_config.num_attention_heads,
            init_method_std=hf_config.initializer_range,
            layernorm_epsilon=hf_config.rms_norm_eps,
            num_query_groups=num_query_groups,
            seq_length=hf_config.max_position_embeddings,
            rotary_base=hf_config.rope_theta,
            kv_channels=getattr(hf_config, "head_dim", None),
            gated_linear_unit=True,  # Llama uses SwiGLU
            make_vocab_size_divisible_by=self.make_vocab_size_divisible_by(hf_config.vocab_size),
            share_embeddings_and_output_weights=getattr(hf_config, "tie_word_embeddings", False),
            fp16=(self.dtype_from_hf(hf_config, default=torch.float32) == torch.float16),
            bf16=(self.dtype_from_hf(hf_config, default=torch.float32) == torch.bfloat16),
            params_dtype=self.dtype_from_hf(hf_config, default=torch.float32),
            vocab_size=hf_config.vocab_size,
        )

        # Handle rope scaling for Llama 3.1/3.3
        if hasattr(hf_config, "rope_scaling") and hf_config.rope_scaling:
            if hf_config.rope_scaling.get("rope_type") == "llama3":
                provider_kwargs["rope_scaling_factor"] = hf_config.rope_scaling.get("factor", 8.0)

        provider_kwargs["heterogeneous_layers_config_encoded_json"] = hf_config.to_json_string()
        provider = LlamaNemotronHeterogeneousProvider(**provider_kwargs)
        return provider

    def mapping_registry(self) -> MegatronMappingRegistry:
        # Return MegatronMappingRegistry containing parameter mappings from Megatron to HF format
        # Similar to Llama bridge but adapted for Llama-Nemotron specifics

        # Dictionary maps Megatron parameter names -> HF parameter names
        # Supports wildcard (*) patterns for layer-specific parameters
        param_mappings = {
            "embedding.word_embeddings.weight": "model.embed_tokens.weight",
            "output_layer.weight": "lm_head.weight",
            "decoder.final_layernorm.weight": "model.norm.weight",
            "decoder.layers.*.self_attention.linear_qkv.layer_norm_weight": "model.layers.*.input_layernorm.weight",
            "decoder.layers.*.mlp.linear_fc1.layer_norm_weight": "model.layers.*.post_attention_layernorm.weight",
            "decoder.layers.*.self_attention.linear_proj.weight": "model.layers.*.self_attn.o_proj.weight",
            "decoder.layers.*.mlp.linear_fc2.weight": "model.layers.*.mlp.down_proj.weight",
        }

        mapping_list = []
        # Convert each dictionary entry to AutoMapping(megatron_param, hf_param)
        for megatron_param, hf_param in param_mappings.items():
            mapping_list.append(AutoMapping(megatron_param=megatron_param, hf_param=hf_param))

        # Add special mappings that require parameter concatenation/transformation
        mapping_list.extend(
            [
                # QKV: Combine separate Q, K, V matrices into single QKV matrix
                QKVMapping(
                    megatron_param="decoder.layers.*.self_attention.linear_qkv.weight",
                    q="model.layers.*.self_attn.q_proj.weight",
                    k="model.layers.*.self_attn.k_proj.weight",
                    v="model.layers.*.self_attn.v_proj.weight",
                ),
                # Gated MLP: Combine gate and up projection matrices into single FC1 matrix
                GatedMLPMapping(
                    megatron_param="decoder.layers.*.mlp.linear_fc1.weight",
                    gate="model.layers.*.mlp.gate_proj.weight",
                    up="model.layers.*.mlp.up_proj.weight",
                ),
            ]
        )

        return MegatronMappingRegistry(*mapping_list)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bridge wrapper classes for Megatron Core transformer configurations.

These classes provide deferred post-initialization to support the Bridge configuration
override system while maintaining compatibility with Megatron Core's post_init behavior.
"""

import copy
from dataclasses import dataclass, fields, is_dataclass

from megatron.core.transformer.heterogeneous.heterogeneous_config import (
    HeterogeneousTransformerConfig as MCoreHeterogeneousTransformerConfig,
)
from megatron.core.transformer.transformer_config import MLATransformerConfig as MCoreMLATransformerConfig
from megatron.core.transformer.transformer_config import TransformerConfig as MCoreTransformerConfig


def _safe_asdict(obj, skip_keys: set[str]) -> dict:
    """Shallow asdict variant that preserves handles like process groups.

    dataclasses.asdict performs a deep copy of every leaf value, which breaks objects that
    should remain shared references (e.g., ProcessGroupCollection). This helper mirrors the
    structure of asdict but returns leaf objects as-is so they are not deep-copied.
    """
    if is_dataclass(obj):
        result = {}
        for f in fields(obj):
            value = getattr(obj, f.name)
            result[f.name] = value if f.name in skip_keys else _safe_asdict(value, skip_keys)
        return result
    if isinstance(obj, (list, tuple)):
        return obj.__class__(_safe_asdict(v, skip_keys) for v in obj)
    if isinstance(obj, dict):
        return obj.__class__((_safe_asdict(k, skip_keys), _safe_asdict(v, skip_keys)) for k, v in obj.items())
    return obj


@dataclass
class TransformerConfig(MCoreTransformerConfig):
    """Megatron Core TransformerConfig with deferred post-init.

    This class inherits from Megatron Core's TransformerConfig but defers the
    execution of post_init() until finalize() is explicitly called. This allows
    for field modifications after construction but before computed fields are
    calculated.

    Usage:
        # Create config with deferred post-init
        config = TransformerConfig(num_layers=32, hidden_size=4096)

        # Modify fields as needed
        config.seq_length = 8192
        config.tensor_model_parallel_size = 2

        # Finalize to compute derived fields
        config.finalize()
    """

    _NO_COPY_KEYS = {"_pg_collection"}

    def __post_init__(self) -> None:
        """Skip MCore post_init during initial construction.

        The original post_init logic is deferred until finalize() is called.
        This allows for field modifications after construction without
        invalidating computed fields.
        """
        pass

    def finalize(self) -> None:
        """Execute the deferred MCore post-init logic.

        This method calls the original Megatron Core TransformerConfig.__post_init__()
        to compute derived fields based on the current field values. It can be
        called multiple times safely.
        """
        if self.pipeline_model_parallel_size > 1 and self.pipeline_dtype is None:
            self.pipeline_dtype = self.params_dtype
        MCoreTransformerConfig.__post_init__(self)

    def __deepcopy__(self, memo):
        """Custom deepcopy to preserve process group handles when cloning configs.

        Certain attributes (_pg_collection, etc.) should remain shared references
        rather than being wiped or re-created during deepcopy.
        TODO: This is a temporary hack. Once providers stop embedding the Transformer
        config and instead hold the MCore config as an attribute, we can remove this.
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for key, value in self.__dict__.items():
            if key in self._NO_COPY_KEYS:
                # Keep the same reference to avoid losing initialized process groups.
                setattr(result, key, value)
            else:
                setattr(result, key, copy.deepcopy(value, memo))
        return result

    def asdict(self) -> dict:
        """Return a dict view without deep-copying shared handles (e.g., process groups)."""
        return _safe_asdict(self, self._NO_COPY_KEYS)


@dataclass
class MLATransformerConfig(TransformerConfig, MCoreMLATransformerConfig):
    """Megatron Core MLATransformerConfig with deferred post-init.

    This class inherits from Megatron Core's MLATransformerConfig but defers the
    execution of post_init() until finalize() is explicitly called. This allows
    for field modifications after construction but before computed fields are
    calculated.

    Usage:
        # Create config with deferred post-init
        config = MLATransformerConfig(num_layers=32, hidden_size=4096)

        # Modify fields as needed
        config.q_lora_rank = 1536
        config.kv_lora_rank = 512

        # Finalize to compute derived fields
        config.finalize()
    """

    def __post_init__(self) -> None:
        """Skip MCore post_init during initial construction.

        The original post_init logic is deferred until finalize() is called.
        This allows for field modifications after construction without
        invalidating computed fields.
        """
        pass

    def finalize(self) -> None:
        """Execute the deferred MCore post-init logic.

        This method calls the original Megatron Core MLATransformerConfig.__post_init__()
        to compute derived fields based on the current field values. It can be
        called multiple times safely.
        """
        if self.pipeline_model_parallel_size > 1 and self.pipeline_dtype is None:
            self.pipeline_dtype = self.params_dtype
        MCoreMLATransformerConfig.__post_init__(self)


@dataclass
class HeterogeneousTransformerConfig(TransformerConfig, MCoreHeterogeneousTransformerConfig):
    """Megatron Core HeterogeneousTransformerConfig with deferred post-init.

    This class inherits from both our lazy TransformerConfig and Megatron Core's
    HeterogeneousTransformerConfig. The MRO ensures that our lazy post-init behavior
    is preserved while maintaining all heterogeneous functionality.

    CRITICAL: The inheritance order is important for MRO:
    1. TransformerConfig (our lazy version) comes first
    2. MCoreHeterogeneousTransformerConfig comes second

    Usage:
        # Create config with deferred post-init
        config = HeterogeneousTransformerConfig(
            num_layers=32,
            hidden_size=4096,
            heterogeneous_layers_config_encoded_json=json_string
        )

        # Modify fields as needed
        config.seq_length = 8192
        config.tensor_model_parallel_size = 2

        # Finalize to compute derived fields and parse heterogeneous config
        config.finalize()
    """

    def __post_init__(self) -> None:
        """Skip MCore post_init during initial construction.

        The original post_init logic is deferred until finalize() is called.
        This allows for field modifications after construction without
        invalidating computed fields.
        """
        pass

    def finalize(self) -> None:
        """Execute the deferred MCore post-init logic.

        This method calls the original Megatron Core HeterogeneousTransformerConfig.__post_init__()
        to compute derived fields and parse the heterogeneous block configurations.
        It can be called multiple times safely.
        """
        MCoreHeterogeneousTransformerConfig.__post_init__(self)

    def get_config_for_layer(self, layer_number: int) -> MCoreTransformerConfig:
        """Return a layer-specific TransformerConfig without deep-copying process groups."""
        # TODO: This is a temporary hack; replace once providers hold the MCore config directly.

        layer_idx = layer_number - 1  # layer number starts from 1
        if layer_idx < 0 or layer_idx >= len(self.per_block_parameters):
            raise ValueError(
                f"Invalid layer number: {layer_number}. Should be in range [1, {len(self.per_block_parameters)}]."
            )
        block_config = self.per_block_parameters[layer_idx]

        keys_to_update = {}

        # attention config updates
        if block_config.attention.num_query_groups is not None:
            assert not block_config.attention.replace_with_linear and not block_config.attention.no_op
            keys_to_update["num_query_groups"] = block_config.attention.num_query_groups

        # mlp config updates
        if block_config.mlp.ffn_hidden_size is not None:
            assert not block_config.mlp.replace_with_linear and not block_config.mlp.no_op
            keys_to_update["ffn_hidden_size"] = block_config.mlp.ffn_hidden_size

        transformer_config_dict = self.asdict()

        # remove keys that are not in TransformerConfig
        transformer_config_field_names = {f.name for f in fields(MCoreTransformerConfig)}
        transformer_config_dict = {
            k: v for k, v in transformer_config_dict.items() if k in transformer_config_field_names
        }

        transformer_config_dict.update(keys_to_update)

        return MCoreTransformerConfig(**transformer_config_dict)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Dict, Mapping

import torch
from megatron.core.models.gpt.gpt_layer_specs import get_gpt_decoder_block_spec
from megatron.core.models.gpt.gpt_model import GPTModel

from megatron.bridge.models.conversion.mapping_registry import MegatronMappingRegistry
from megatron.bridge.models.conversion.model_bridge import MegatronModelBridge, WeightConversionTask
from megatron.bridge.models.conversion.param_mapping import AutoMapping
from megatron.bridge.models.deepseek.common import get_common_mapping_list
from megatron.bridge.models.hf_pretrained.causal_lm import PreTrainedCausalLM
from megatron.bridge.models.mla_provider import MLAModelProvider


try:
    import transformer_engine  # noqa: F401

    HAVE_TE = True
except (ImportError, ModuleNotFoundError):
    HAVE_TE = False


@MegatronModelBridge.register_bridge(
    source="DeepseekV3ForCausalLM",
    target=GPTModel,
    provider=MLAModelProvider,
    model_type="deepseek_v3",
)
class DeepSeekV3Bridge(MegatronModelBridge):
    """Megatron Bridge for DeepSeek-V3."""

    def provider_bridge(self, hf_pretrained: PreTrainedCausalLM) -> MLAModelProvider:
        provider = super().provider_bridge(hf_pretrained)
        hf_config = hf_pretrained.config

        provider.transformer_layer_spec = partial(get_gpt_decoder_block_spec, use_transformer_engine=HAVE_TE)
        provider.normalization = "RMSNorm"
        provider.gated_linear_unit = True
        provider.position_embedding_type = "rope"
        provider.add_bias_linear = False
        provider.share_embeddings_and_output_weights = False
        provider.qk_layernorm = True
        provider.multi_latent_attention = True

        provider.moe_grouped_gemm = True
        provider.moe_router_pre_softmax = True
        provider.moe_token_dispatcher_type = "alltoall"
        provider.moe_router_load_balancing_type = "seq_aux_loss"
        provider.moe_shared_expert_overlap = True
        provider.moe_router_enable_expert_bias = True
        provider.moe_router_dtype = "fp32"
        provider.moe_permute_fusion = True
        provider.moe_aux_loss_coeff = 0.0001

        provider.apply_rope_fusion = False
        provider.bias_activation_fusion = True
        provider.bias_dropout_fusion = True
        provider.cross_entropy_fusion_impl = "te"
        provider.cross_entropy_loss_fusion = True
        provider.masked_softmax_fusion = True
        provider.persist_layer_norm = True
        provider.async_tensor_model_parallel_allreduce = True
        provider.gradient_accumulation_fusion = True

        provider.hidden_dropout = 0.0
        provider.attention_softmax_in_fp32 = False

        provider.make_vocab_size_divisible_by = 1280
        provider.seq_length = 4096

        provider.moe_layer_freq = [0] * hf_config.first_k_dense_replace + [1] * (
            hf_config.num_hidden_layers - hf_config.first_k_dense_replace
        )
        provider.moe_shared_expert_intermediate_size = hf_config.moe_intermediate_size * hf_config.n_shared_experts

        # TODO: mtp
        provider.mtp_num_layers = None

        return provider

    def mapping_registry(self) -> MegatronMappingRegistry:
        mapping_list = get_common_mapping_list()
        mapping_list.append(
            AutoMapping(
                megatron_param="decoder.layers.*.mlp.router.expert_bias",
                hf_param="model.layers.*.mlp.gate.e_score_correction_bias",
            )
        )
        return MegatronMappingRegistry(*mapping_list)

    def maybe_modify_converted_hf_weight(
        self,
        task: WeightConversionTask,
        converted_weights_dict: Dict[str, torch.Tensor],
        hf_state_dict: Mapping[str, torch.Tensor],
    ) -> Dict[str, torch.Tensor]:
        """Add rotary embedding inverse frequency parameter if needed."""
        global_name = task.global_param_name
        if not global_name.startswith("decoder.layers.") or not global_name.endswith(".input_layernorm.weight"):
            return converted_weights_dict

        parts = global_name.split(".")
        if len(parts) < 4 or not parts[2].isdigit():
            return converted_weights_dict

        inv_freq_prefix = "model.layers."
        inv_freq_suffix = ".self_attn.rotary_emb.inv_freq"
        layer_idx = int(parts[2])
        inv_freq_key = f"{inv_freq_prefix}{layer_idx}{inv_freq_suffix}"
        if inv_freq_key in converted_weights_dict:
            return converted_weights_dict

        has_inv_freq = getattr(self, "_deepseek_has_inv_freq", None)
        if has_inv_freq is None:
            has_inv_freq = False
            for key in hf_state_dict.keys():
                if key.startswith(inv_freq_prefix) and key.endswith(inv_freq_suffix):
                    has_inv_freq = True
                    break
            self._deepseek_has_inv_freq = has_inv_freq
        if not has_inv_freq:
            return converted_weights_dict

        inv_freq = getattr(self, "_deepseek_inv_freq", None)
        if inv_freq is None:
            rotary_dim = self.hf_config.qk_rope_head_dim
            rotary_base = self.hf_config.rope_theta
            inv_freq = 1.0 / (rotary_base ** (torch.arange(0, rotary_dim, 2, dtype=torch.float32) / rotary_dim))
            self._deepseek_inv_freq = inv_freq

        if converted_weights_dict:
            reference_tensor = next(iter(converted_weights_dict.values()))
            if inv_freq.device != reference_tensor.device:
                inv_freq = inv_freq.to(device=reference_tensor.device)
                self._deepseek_inv_freq = inv_freq

        converted_weights_dict[inv_freq_key] = inv_freq
        return converted_weights_dict



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import dataclass, field
from functools import partial
from typing import TYPE_CHECKING, Callable, List, Optional, Union

import torch
import torch.nn.functional as F
from megatron.core.models.gpt.gpt_layer_specs import get_gpt_decoder_block_spec

from megatron.bridge.models.gpt_provider import GPTModelProvider
from megatron.bridge.models.transformer_config import MLATransformerConfig


try:
    import transformer_engine  # type: ignore  # noqa: F401

    HAVE_TE = True
except (ImportError, ModuleNotFoundError):
    HAVE_TE = False

if TYPE_CHECKING:
    from megatron.core.transformer import ModuleSpec

if HAVE_TE:
    from megatron.core.utils import is_te_min_version


@dataclass
class KimiK2Provider(MLATransformerConfig, GPTModelProvider):
    """
    https://moonshotai.github.io/Kimi-K2/
    """

    transformer_layer_spec: Union["ModuleSpec", Callable[["GPTModelProvider"], "ModuleSpec"]] = partial(
        get_gpt_decoder_block_spec, use_transformer_engine=HAVE_TE
    )

    # Model
    num_layers: int = 61
    hidden_size: int = 7168
    ffn_hidden_size: int = 18432
    num_moe_experts: int = 384
    moe_ffn_hidden_size: int = 2048
    moe_shared_expert_intermediate_size: int = 2048  # 2048 * 1 shared expert
    moe_layer_freq: Union[int, List[int]] = field(default_factory=lambda: [0] + [1] * 60)  # first layer are dense
    normalization: str = "RMSNorm"
    activation_func: Callable = F.silu
    gated_linear_unit: bool = True  # swiglu
    position_embedding_type: str = "rope"
    add_bias_linear: bool = False
    share_embeddings_and_output_weights: bool = False
    num_attention_heads: int = 64
    kv_channels: int = 64
    max_position_embeddings: int = 4096
    seq_length: int = 4096
    rotary_base: float = 50000.0
    make_vocab_size_divisible_by: int = 1280
    mtp_num_layers: Optional[int] = None
    mtp_loss_scaling_factor: Optional[float] = None

    # Regularization
    attention_dropout: float = 0.0
    hidden_dropout: float = 0.0
    qk_layernorm: bool = True

    # MoE
    moe_router_topk: int = 8
    moe_router_num_groups: int = 1
    moe_router_group_topk: int = 1
    moe_router_topk_scaling_factor: float = 2.827
    moe_aux_loss_coeff: float = 1e-3
    moe_router_score_function: str = "sigmoid"
    moe_router_enable_expert_bias: bool = True
    moe_router_bias_update_rate: float = 1e-3
    moe_grouped_gemm: bool = True
    moe_router_pre_softmax: bool = True
    moe_token_dispatcher_type: str = "alltoall"
    moe_router_load_balancing_type: str = "seq_aux_loss"
    moe_shared_expert_overlap: bool = True
    moe_router_dtype: Optional[str] = "fp32"

    # MLA
    multi_latent_attention: bool = True
    q_lora_rank: int = 1536
    kv_lora_rank: int = 512
    qk_head_dim: int = 128
    qk_pos_emb_head_dim: int = 64
    v_head_dim: int = 128
    rotary_scaling_factor: float = 32
    beta_fast: float = 1.0
    beta_slow: float = 1.0
    mscale: float = 1.0
    mscale_all_dim: float = 1.0

    # Miscellaneous
    init_method_std: float = 0.006
    layernorm_epsilon: float = 1e-6
    bf16: bool = True
    params_dtype: torch.dtype = torch.bfloat16
    async_tensor_model_parallel_allreduce: bool = True
    attention_softmax_in_fp32: bool = False
    persist_layer_norm: bool = True
    num_layers_in_first_pipeline_stage: Optional[int] = None
    num_layers_in_last_pipeline_stage: Optional[int] = None
    account_for_embedding_in_pipeline_split: bool = False
    account_for_loss_in_pipeline_split: bool = False
    vocab_size: int = 163840

    # fusions
    apply_rope_fusion: bool = False
    bias_activation_fusion: bool = True
    bias_dropout_fusion: bool = True
    masked_softmax_fusion: bool = True
    gradient_accumulation_fusion: bool = True
    cross_entropy_loss_fusion: bool = True
    cross_entropy_fusion_impl: str = "te"
    moe_permute_fusion: bool = is_te_min_version("2.1.0") if HAVE_TE else False



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
GLM 4.5 Vision-Language Model for Megatron.

This module provides the GLM45VLModel class that combines:
- HuggingFace's vision encoder (vision_tower) for image processing
- Megatron's language model for text generation

Reference: https://huggingface.co/zai-org/GLM-4.5V
"""

import types
from typing import TYPE_CHECKING, Optional

import torch
import transformers
from megatron.core.tensor_parallel import scatter_to_sequence_parallel_region
from megatron.core.transformer.module import MegatronModule
from packaging.version import Version as PkgVersion
from torch import Tensor
from transformers.models.glm4v.modeling_glm4v import Glm4vModel

from megatron.bridge.models.gpt_provider import GPTModelProvider
from megatron.bridge.utils.common_utils import hook_hf_module_setattr_for_tp_grad_sync


if TYPE_CHECKING:
    from megatron.core.packed_seq_params import PackedSeqParams


def is_transformers_min_version(version):
    """Check if minimum version of transformers is installed."""
    try:
        transformers_version = PkgVersion(transformers.__version__)
        return transformers_version >= PkgVersion(version)
    except Exception:
        # If version parsing fails, assume false for safety
        return False


class GLM45VModel(MegatronModule):
    """
    GLM 4.5 Vision-Language (VL) model wrapper for Megatron.

    This class combines HuggingFace's vision components with Megatron's language model:
    - Vision tower (HF): Processes images through the vision encoder
    - Multimodal projector (HF): Projects vision features to language model space
    - Language model (Megatron): Generates text conditioned on vision and text inputs

    The vision encoder forward pass uses HuggingFace implementation via monkey-patching,
    while the language model forward pass uses Megatron's optimized implementation.

    Args:
        config (GPTModelProvider): Model provider containing configuration for language and vision modules.
        pre_process (bool, optional): Whether to construct the vision tower and projector. Default: True.
        post_process (bool, optional): Whether to apply post-processing. Default: True.
        vp_stage (Optional[int], optional): Pipeline stage for model parallelism. Default: None.

    Attributes:
        pre_process (bool): If True, enables vision and multimodal components.
        post_process (bool): If True, enables post-processing.
        vp_stage (Optional[int]): Pipeline stage for model parallelism.
        vision_tower (nn.Module): Vision encoder from HuggingFace.
        multi_modal_projector (nn.Module): Projects vision features to language model space.
        language_model (nn.Module): Megatron language model.
        get_image_features (callable): Method to extract image features (monkey-patched from HF).

    Forward Inputs:
        input_ids (torch.LongTensor, optional): Tokenized input ids for the language model.
        attention_mask (torch.Tensor, optional): Attention mask for the language model.
        position_ids (torch.LongTensor, optional): Position ids for the language model.
        inputs_embeds (torch.FloatTensor, optional): Precomputed input embeddings.
        pixel_values (torch.Tensor, optional): Image tensor(s) for the vision tower.
        labels (torch.Tensor, optional): Target labels for supervised training.
        runtime_gather_output (bool, optional): If True, gather outputs across pipeline stages.
        loss_mask (Tensor, optional): Mask for loss computation.

    Returns:
        Tensor: Model output (e.g., logits or loss, depending on mode).

    Note:
        - If `pre_process` is False, only the language model is constructed.
        - The vision tower and projector are only active if `pre_process` is True.
        - This class is intended for use within the Megatron-LM framework.
        - Requires transformers >= 5.0.0 for Mistral3 model support.
    """

    def __init__(
        self,
        config: GPTModelProvider,
        pre_process: bool = True,
        post_process: bool = True,
        vp_stage: Optional[int] = None,
    ) -> None:
        super().__init__(config=config)

        self.pre_process = pre_process
        self.post_process = post_process
        self.vp_stage = vp_stage

        # Bind methods from HF's Qwen2_5_VLModel to this instance
        # get_placeholder_mask is only available in transformers 4.55+
        if not is_transformers_min_version("4.57.1"):
            raise RuntimeError(
                f"transformers version {transformers.__version__} is not supported. "
                f"get_placeholder_mask requires transformers >= 4.55.0. "
                f"Please upgrade transformers: pip install 'transformers>=4.55.0'"
            )

        if pre_process:
            from transformers.models.glm4v.modeling_glm4v import Glm4vVisionModel

            self.visual = Glm4vVisionModel._from_config(config.vision_config)
            # Ensure HF visual tower params are marked for TP grad sync
            hook_hf_module_setattr_for_tp_grad_sync(self.visual)

        # Initialize Megatron language model
        self.language_model = self.config.provide_language_model(
            pre_process=pre_process, post_process=post_process, vp_stage=vp_stage
        )

        # Finalize grad requires these to be bound with module
        self.share_embeddings_and_output_weights = config.share_embeddings_and_output_weights
        self.shared_embedding_or_output_weight = self.language_model.shared_embedding_or_output_weight

        # Monkey-patch methods from HuggingFace Mistral3Model
        # This allows us to use HF's image feature extraction logic
        self.get_image_features = types.MethodType(Glm4vModel.get_image_features, self)
        self.get_video_features = types.MethodType(Glm4vModel.get_video_features, self)
        self.get_rope_index = types.MethodType(Glm4vModel.get_rope_index, self)
        self.get_placeholder_mask = types.MethodType(Glm4vModel.get_placeholder_mask, self)

        # Some config requires from HF vision tower
        self.config.spatial_merge_size = getattr(self.config.vision_config, "spatial_merge_size", 2)

    def set_input_tensor(self, input_tensor) -> None:
        """Set model chunk input tensor."""
        self.language_model.set_input_tensor(input_tensor)

    def forward(
        self,
        input_ids: torch.LongTensor = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        pixel_values: Optional[torch.Tensor] = None,
        pixel_values_videos: Optional[torch.FloatTensor] = None,
        image_grid_thw: Optional[torch.LongTensor] = None,
        video_grid_thw: Optional[torch.LongTensor] = None,
        labels: Optional[torch.Tensor] = None,
        runtime_gather_output: Optional[bool] = None,
        packed_seq_params: Optional["PackedSeqParams"] = None,
        *,
        loss_mask: Optional[Tensor] = None,
    ) -> Tensor:
        """
        Forward pass combining HuggingFace vision encoder with Megatron language model.

        Args:
            input_ids: Tokenized input ids for the language model.
            attention_mask: Attention mask for the language model.
            position_ids: Position ids for the language model.
            inputs_embeds: Precomputed input embeddings.
            pixel_values: Image tensor(s) for the vision tower.
            pixel_values_videos: Video tensor(s) for the vision tower.
            image_grid_thw: Grid of image sizes for the vision tower.
            video_grid_thw: Grid of video sizes for the vision tower.
            second_per_grid_ts: Time interval for each grid along the temporal dimension in the 3D position IDs.
            labels: Target labels for supervised training.
            runtime_gather_output: If True, gather outputs across pipeline stages.
            loss_mask: Mask for loss computation.

        Returns:
            Model output (logits or loss depending on mode).
        """
        if self.pre_process:
            if inputs_embeds is None:
                # Get text embeddings from Megatron language model
                inputs_embeds = self.language_model.embedding(
                    input_ids=input_ids, position_ids=None
                )  # [seq_len, batch, hidden]

                # Transpose to HF format [batch, seq_len, hidden]
                inputs_embeds = inputs_embeds.transpose(1, 0).contiguous()

            if pixel_values is not None:
                image_embeds = self.get_image_features(pixel_values, image_grid_thw)
                image_embeds = torch.cat(image_embeds, dim=0).to(inputs_embeds.device, inputs_embeds.dtype)
                image_mask, _ = self.get_placeholder_mask(
                    input_ids, inputs_embeds=inputs_embeds, image_features=image_embeds
                )
                inputs_embeds = inputs_embeds.masked_scatter(image_mask, image_embeds)

            if pixel_values_videos is not None:
                video_embeds = self.get_video_features(pixel_values_videos, video_grid_thw)
                video_embeds = torch.cat(video_embeds, dim=0).to(inputs_embeds.device, inputs_embeds.dtype)
                _, video_mask = self.get_placeholder_mask(
                    input_ids, inputs_embeds=inputs_embeds, video_features=video_embeds
                )
                inputs_embeds = inputs_embeds.masked_scatter(video_mask, video_embeds)

            # Transpose back to Megatron format [seq_len, batch, hidden]
            inputs_embeds = inputs_embeds.transpose(1, 0).contiguous()

            if self.config.sequence_parallel:
                inputs_embeds = scatter_to_sequence_parallel_region(inputs_embeds)

        # Compute MRoPE position_ids on ALL pipeline stages
        # Each stage has input_ids and visual grid info from the data iterator
        # This avoids any broadcasting overhead
        hf_attention_mask = None
        position_ids, rope_deltas = self.get_rope_index(
            input_ids,
            image_grid_thw,
            video_grid_thw,
            attention_mask=hf_attention_mask,
        )

        # Forward through Megatron language model
        outputs = self.language_model.forward(
            input_ids=None,
            position_ids=position_ids,
            attention_mask=attention_mask,
            decoder_input=inputs_embeds,
            labels=labels,
            loss_mask=loss_mask,
            runtime_gather_output=runtime_gather_output,
            packed_seq_params=packed_seq_params,
        )
        return outputs

    def freeze(
        self,
        freeze_language_model: bool,
        freeze_vision_model: bool,
        freeze_vision_projection: bool,
    ):
        """Freeze model modules.

        Make specific modules non-trainable by setting requires_grad to False.

        Args:
            freeze_language_model (bool): Freeze the language model module.
            freeze_vision_model (bool): Freeze the vision model module (patch_embed and blocks).
            freeze_vision_projection (bool): Freeze the vision projection module (merger).
        """
        modules = []

        if freeze_language_model and hasattr(self, "language_model") and self.language_model is not None:
            modules.append(self.language_model)

        if freeze_vision_model and hasattr(self, "visual") and self.visual is not None:
            # Vision model consists of patch_embed and blocks
            if hasattr(self.visual, "patch_embed"):
                modules.append(self.visual.patch_embed)
            if hasattr(self.visual, "blocks"):
                modules.append(self.visual.blocks)

        if freeze_vision_projection and hasattr(self, "visual") and self.visual is not None:
            # Vision projection is the merger module
            if hasattr(self.visual, "merger"):
                modules.append(self.visual.merger)

        for module in modules:
            for param in module.parameters():
                param.requires_grad = False



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from megatron.bridge.models.distillation_provider import DistillationProvider
from megatron.bridge.training.config import ConfigContainer
from megatron.bridge.training.gpt_step import forward_step_modelopt
from megatron.bridge.training.pretrain import pretrain
from megatron.bridge.utils.decorators import experimental_fn


@experimental_fn
def distill(
    config: ConfigContainer,
) -> None:
    """Main function to run knowledge distillation (KD).

    Args:
        config: The main configuration container holding all necessary parameters.

    Warnings:
        This is an experimental API and is subject to change in backwards
        incompatible ways without notice.
    """
    assert isinstance(config.model, DistillationProvider), "Distillation requires a DistillationProvider"

    return pretrain(config, forward_step_modelopt)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Built-in maker functions that transform HuggingFace datasets into
conversation-style examples consumable by VLM processors.
"""

import json
import random
from pathlib import Path
from typing import Any, Dict, List

from datasets import concatenate_datasets, load_dataset

from megatron.bridge.data.vlm_datasets.token_utils import json2token
from megatron.bridge.utils.common_utils import resolve_path


def make_rdr_dataset(
    path_or_dataset: str = "quintend/rdr-items", split: str = "train", **kwargs
) -> List[Dict[str, Any]]:
    """Load and preprocess the RDR dataset for image-to-text fine-tuning.

    Returns a list of examples with a "conversation" field that includes an image and text.
    """
    dataset = load_dataset(path_or_dataset, split=split)

    def format(example):
        return {
            "conversation": [
                {
                    "role": "user",
                    "content": [
                        {"type": "image", "image": example["image"]},
                        {"type": "text", "text": "Describe this image."},
                    ],
                },
                {
                    "role": "assistant",
                    "content": [{"type": "text", "text": example["text"]}],
                },
            ],
        }

    return [format(example) for example in dataset]


def make_cord_v2_dataset(
    path_or_dataset: str = "naver-clova-ix/cord-v2", split: str = "train", **kwargs
) -> List[Dict[str, Any]]:
    """Load and preprocess the CORD-V2 dataset for image-to-text fine-tuning."""
    dataset = load_dataset(path_or_dataset, split=split)

    def format(example):
        ground_truth = json.loads(example["ground_truth"])
        if "gt_parses" in ground_truth:
            assert isinstance(ground_truth["gt_parses"], list)
            gt_jsons = ground_truth["gt_parses"]
        else:
            assert "gt_parse" in ground_truth and isinstance(ground_truth["gt_parse"], dict)
            gt_jsons = [ground_truth["gt_parse"]]

        text = random.choice([json2token(gt_json, sort_json_key=True) for gt_json in gt_jsons])

        return {
            "conversation": [
                {
                    "role": "user",
                    "content": [
                        {"type": "image", "image": example["image"]},
                        {"type": "text", "text": "Describe this image."},
                    ],
                },
                {"role": "assistant", "content": [{"type": "text", "text": text}]},
            ],
        }

    return [format(example) for example in dataset]


def make_medpix_dataset(
    path_or_dataset: str = "mmoukouba/MedPix-VQA", split: str = "train", **kwargs
) -> List[Dict[str, Any]]:
    """Load and preprocess the MedPix dataset for image-to-text fine-tuning."""
    dataset = load_dataset(path_or_dataset, split=split)

    def format(example):
        return {
            "conversation": [
                {
                    "role": "user",
                    "content": [
                        {"type": "image", "image": example["image_id"]},
                        {"type": "text", "text": example["question"]},
                    ],
                },
                {"role": "assistant", "content": [{"type": "text", "text": example["answer"]}]},
            ],
        }

    return [format(example) for example in dataset]


def make_raven_dataset(
    path_or_dataset: str = "HuggingFaceM4/the_cauldron",
    subset: str = "raven",
    split: str = "train",
    **kwargs,
) -> List[Dict[str, Any]]:
    """Load and preprocess the Raven subset from the Cauldron dataset.

    This subset follows the IDEFICS-style layout where each sample contains:
    - ``images``: a (possibly empty) list of PIL images
    - ``texts``: a list of conversation dictionaries. For Raven, ``texts[0]``
      is a *single* turn stored as a dictionary with two keys::

          {"user": "<question>", "assistant": "<answer>"}

      Only the first element is used.  The ``user`` string is taken as the
      user prompt, and ``assistant`` is the ground-truth answer.

    Conversation building policy:
    1. All images are placed at the beginning of the user turn followed by the
       textual prompt.
    2. The assistant turn contains the answer text.

    Examples missing either images or the required fields are filtered out.
    """
    if split != "train":
        raise ValueError("Raven dataset only supports train split. Please set `train.eval_iters=0`.")
    dataset = load_dataset(path_or_dataset, subset, split=split)

    def format(example):
        images = example.get("images", [])
        texts = example.get("texts", [])
        if not images or not texts or not isinstance(texts[0], dict):
            return None

        user_prompt = texts[0].get("user")
        assistant_answer = texts[0].get("assistant")
        if user_prompt is None or assistant_answer is None:
            return None

        user_content: List[Dict[str, Any]] = [{"type": "image", "image": img} for img in images]
        user_content.append({"type": "text", "text": user_prompt})

        assistant_content = [{"type": "text", "text": assistant_answer}]

        return {
            "conversation": [
                {"role": "user", "content": user_content},
                {"role": "assistant", "content": assistant_content},
            ]
        }

    formatted = (format(example) for example in dataset)
    # Filter out any None values from malformed rows.
    return [ex for ex in formatted if ex is not None]


def make_llava_video_178k_dataset(
    video_root_path: str,
    path_or_dataset: str = "lmms-lab/LLaVA-Video-178K",
    subsets: str | List[str] = "0_30_s_nextqa",
    split: str = "open_ended",
) -> List[Dict[str, Any]]:
    """Load and preprocess a subset of the *LLaVA-Video-178K* dataset.

    Each row contains:
    - ``video``: path or URL to the MP4 file.
    - ``conversations``: a **two-turn** list::

          [{"from": "human", "value": "<video>\n<question>"},
           {"from": "gpt",   "value": "<answer>"}]

      We map this schema to our internal multimodal conversation format:

      User turn  →  [video, user prompt]
      Assistant  →  answer text

    Note:
        Video files are assumed to be pre-downloaded and stored locally in the
        ``video_root_path`` directory. Rows with missing videos or empty
        conversations are filtered out from the final output.

    Args:
        video_root_path: Root directory where video files are stored locally.
        path_or_dataset: HF dataset path or local cache dir.
        subsets: Single subset name or list of the dataset's directory-style
            subsets to load.
        split: Split to load from the dataset. Note that "train" is automatically
            mapped to "open_ended".

    Returns:
        A list of dicts each containing a ``conversation`` field ready for
        downstream VLM processors.
    """
    if isinstance(subsets, str):
        subsets = [subsets]

    if split == "train":
        split = "open_ended"
    elif split in ("validation", "test"):
        raise ValueError("LLaVA-Video-178K dataset only supports train split. Please set `train.eval_iters=0`.")
    individual_datasets = [load_dataset(path_or_dataset, subset, split=split) for subset in subsets]
    dataset = concatenate_datasets(individual_datasets)

    # FIXME: right now we assume the video files are pre-downloaded and stored in the video_root_path
    # we need to modify this to download the video files from the hub if they are not present in the video_root_path

    def clean_prompt(val: str) -> str:
        # Remove placeholder tokens such as <image> or <video>
        val = val.replace("<image>", "").replace("<video>", "").strip()
        return val.lstrip("\n").rstrip()

    def format(example):
        video = example.get("video")
        convs = example.get("conversations", [])
        if video in (None, "") or not convs:
            return None

        conversation: List[Dict[str, Any]] = []

        first_human_handled = False
        for turn in convs:
            role = turn.get("from")
            value = turn.get("value", "")
            if not value:
                continue
            if role == "human":
                content: List[Dict[str, Any]] = []
                if not first_human_handled:
                    abs_path = resolve_path(Path(video_root_path) / video)
                    content.append({"type": "video", "path": str(abs_path)})
                    first_human_handled = True
                content.append({"type": "text", "text": clean_prompt(value)})
                conversation.append({"role": "user", "content": content})
            elif role == "gpt":
                conversation.append({"role": "assistant", "content": [{"type": "text", "text": value.strip()}]})

        if not conversation:
            return None

        return {"conversation": conversation}

    formatted = (format(ex) for ex in dataset)
    return [ex for ex in formatted if ex is not None]


def make_cv17_dataset(
    path_or_dataset: str = "ysdede/commonvoice_17_tr_fixed", split: str = "train", **kwargs
) -> List[Dict[str, Any]]:
    """Load and preprocess the CommonVoice 17 dataset for audio-to-text fine-tuning."""
    dataset = load_dataset(path_or_dataset, split=split)
    # Be robust to simple list-like datasets used in tests without `column_names` attr
    try:
        all_columns = dataset.column_names  # type: ignore[attr-defined]
    except Exception:
        first_example = dataset[0] if len(dataset) > 0 else {}
        all_columns = list(first_example.keys()) if isinstance(first_example, dict) else []
    if hasattr(dataset, "remove_columns"):
        columns_to_remove = [col for col in all_columns if col not in ["audio", "transcription"]]
        dataset = dataset.remove_columns(columns_to_remove)

    def format(example):
        return {
            "conversation": [
                {"role": "user", "content": "<|audio_1|>Transcribe the Turkish audio clip."},
                {"role": "assistant", "content": example["transcription"]},
            ],
            "audio": (example["audio"]["array"], example["audio"]["sampling_rate"]),
        }

    return [format(example) for example in dataset]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Core dataset types for conversation-style VLM examples.
"""

from typing import Any, Callable, Dict, List, Optional

import torch

from megatron.bridge.data.vlm_datasets.collate import COLLATE_FNS


class VLMConversationDataset(torch.utils.data.Dataset):
    """Repeating wrapper over a list of HF-style conversation examples.

    - Each base example is expected to contain a "conversation" key following
      processor.apply_chat_template conventions. Optional modality fields like
      "audio" are passed through and consumed by the collate function.
    - Dataset length is set to a target length and indexes wrap around the
      underlying list to meet the requested size.
    - A `collate_fn` attribute is exposed so the framework can pass it to the
      DataLoader.
    """

    def __init__(
        self,
        base_examples: List[Dict[str, Any]],
        target_length: int,
        processor: Any,
        collate_impl: Optional[Callable[[list, Any], Dict[str, torch.Tensor]]] = None,
    ) -> None:
        assert isinstance(base_examples, list) and len(base_examples) > 0, "base_examples must be a non-empty list"
        self._base_examples = base_examples
        self._length = int(max(0, target_length))
        self._processor = processor
        # Choose collate implementation by processor type name when not provided
        collate_key = type(processor).__name__ if processor is not None else "default"
        selected_impl = collate_impl or COLLATE_FNS.get(collate_key, COLLATE_FNS["default"])  # type: ignore[index]

        def _bound_collate(batch: list) -> Dict[str, torch.Tensor]:
            return selected_impl(batch, self._processor)  # type: ignore[call-arg]

        self.collate_fn = _bound_collate

    def __len__(self) -> int:
        return self._length

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        if self._length == 0:
            raise IndexError("Empty dataset")
        base = self._base_examples[idx % len(self._base_examples)]
        return base



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import os
import socket
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Generator, Literal, Optional, Union

import torch
import torch.distributed as dist
from megatron.core import parallel_state
from megatron.core.optimizer import OptimizerConfig
from megatron.core.transformer import MegatronModule, TransformerConfig
from megatron.core.utils import get_model_config

from megatron.bridge.models.model_provider import ModelParallelKwargs, ModelProviderMixin
from megatron.bridge.training.checkpointing import save_checkpoint
from megatron.bridge.training.config import CheckpointConfig, ConfigContainer, LoggerConfig
from megatron.bridge.training.state import GlobalState
from megatron.bridge.training.tokenizers.tokenizer import MegatronTokenizer, build_tokenizer
from megatron.bridge.training.utils.checkpoint_utils import file_exists
from megatron.bridge.utils.vocab_utils import calculate_padded_vocab_size


logger = logging.getLogger(__name__)

HF_BASED_TOKENIZERS = [
    "BertWordPieceLowerCase",
    "BertWordPieceCase",
    "GPT2BPETokenizer",
    "HuggingFaceTokenizer",
]


def torch_dtype_from_mcore_config(config: Any) -> torch.dtype:
    """Convert Megatron-Core config dtype settings to torch dtype.

    Args:
        config: Megatron-Core configuration object with bf16/fp16 flags.

    Returns:
        The corresponding torch dtype.
    """
    if hasattr(config, "bf16") and config.bf16:
        return torch.bfloat16
    elif hasattr(config, "fp16") and config.fp16:
        return torch.float16
    else:
        return torch.float32


@contextmanager
def megatron_cpu_init_context(config: Any) -> Generator[None, None, None]:
    """Context manager to temporarily force CPU initialization for Megatron models.

    This is useful when initializing a model on a system without GPUs or when
    memory constraints prevent GPU initialization.

    Args:
        config: The Megatron model configuration object (e.g., GPTConfig).
            Must have a `use_cpu_initialization` attribute.

    Yields:
        None. The context modifies the config in place.
    """
    original_use_cpu_initialization = config.use_cpu_initialization
    config.use_cpu_initialization = True

    try:
        yield
    finally:
        config.use_cpu_initialization = original_use_cpu_initialization


@contextmanager
def temporary_distributed_context(backend: str = "gloo") -> Generator[None, None, None]:
    """Context manager to temporarily initialize a minimal distributed environment.

    Sets up a single-process distributed backend, initializes Megatron model parallel state,
    yields control, and then cleans up the distributed environment.
    Useful for operations that require Megatron's parallel state but should run
    standalone (e.g., loading distributed checkpoints).

    Args:
        backend: The distributed backend to use ("gloo" for CPU, "nccl" for GPU).

    Yields:
        None.
    """
    if "MASTER_ADDR" in os.environ and "MASTER_PORT" in os.environ:
        init_method = None
    else:
        # Find an available port dynamically
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(("localhost", 0))
            addr, port = s.getsockname()
        init_method = f"tcp://{addr}:{port}"

    dist.init_process_group(backend=backend, init_method=init_method, world_size=1, rank=0)
    parallel_state.initialize_model_parallel()

    # Initialize RNG tracker for model initialization
    # This is needed even for CPU/gloo backend as some model components
    # (like MambaMixer) may use get_cuda_rng_tracker().fork() during __init__
    if torch.cuda.is_available() and torch.cuda.device_count() > 0:
        try:
            from megatron.core.tensor_parallel import model_parallel_cuda_manual_seed

            model_parallel_cuda_manual_seed(0)
        except Exception:
            # If RNG tracker initialization fails (e.g., in test environments
            # or when parallel groups aren't fully initialized), continue anyway.
            # Models that need the RNG tracker will fail later with a clearer error.
            pass

    try:
        yield
    finally:
        parallel_state.destroy_model_parallel()
        dist.destroy_process_group()


def load_tokenizer(checkpoint_path: str, **kwargs) -> MegatronTokenizer:
    """Create a tokenizer from a training checkpoint.

    Obtains tokenizer configuration from the checkpoint and builds the tokenizer.
    Checkpoint should be in MCore distributed checkpoint format.

    Args:
        checkpoint_path: path to an MCore distributed checkpoint directory
                          (e.g., /path/to/model/checkpoints/iter_0000001).
        **kwargs: Overrides to the TokenizerConfig used to build the tokenizer.
                Useful if the tokenizer assets have moved since training.

    """
    from megatron.bridge.training.checkpointing import (
        get_checkpoint_run_config_filename,
        read_run_config,
    )
    from megatron.bridge.training.mlm_compat.arguments import _load_args_from_checkpoint, _tokenizer_config_from_args
    from megatron.bridge.utils.instantiate_utils import instantiate

    run_config_filename = get_checkpoint_run_config_filename(checkpoint_path)

    if file_exists(run_config_filename):
        run_config = read_run_config(run_config_filename)
        mbridge_ckpt = True
    else:
        try:
            mlm_args = _load_args_from_checkpoint(checkpoint_path)
            mbridge_ckpt = False
        except AssertionError:
            raise RuntimeError(f"Checkpoint at {checkpoint_path} is not in a supported format.")

    if mbridge_ckpt:
        cfg = instantiate(run_config["tokenizer"])
    else:
        cfg = _tokenizer_config_from_args(mlm_args)

    for key, val in kwargs.items():
        if hasattr(cfg, key):
            setattr(cfg, key, val)
        else:
            raise AttributeError(
                f"Attempting to set a non-existent attribute '{key}' on TokenizerConfig.\nState of TokenizerConfig before attempting this override: {cfg}"
            )

    if cfg.tokenizer_type in HF_BASED_TOKENIZERS and cfg.tokenizer_model == Path():
        cfg.tokenizer_model = Path(checkpoint_path) / "tokenizer"

    return build_tokenizer(cfg)


def load_model_config(
    checkpoint_path: str,
) -> tuple[TransformerConfig, Optional[argparse.Namespace]]:
    """Returns the model config saved in the checkpoint.

    Supports checkpoints saved with either Megatron Bridge or MegatronLM.

    Args:
        checkpoint_path: path to an MCore distributed checkpoint directory
                          (e.g., /path/to/model/checkpoints/iter_0000001).

    Returns:
        - The model config from the checkpoint. The object returned will be a
          model provider (e.g. GPTModelProvider) if using a Megatron Bridge
          checkpoint or a TransformerConfig if using a MegatronLM checkpoint.
        - If the checkpoint is from MegatronLM, returns the argparse.Namespace
          object. Otherwise None.
    """
    from megatron.bridge.training.checkpointing import (
        get_checkpoint_run_config_filename,
        read_run_config,
    )
    from megatron.bridge.training.mlm_compat.arguments import _load_args_from_checkpoint, _transformer_config_from_args
    from megatron.bridge.utils.instantiate_utils import instantiate

    run_config_filename = get_checkpoint_run_config_filename(checkpoint_path)

    if file_exists(run_config_filename):
        run_config = read_run_config(run_config_filename)
        mbridge_ckpt = True
        mlm_args = None
    else:
        try:
            mlm_args = _load_args_from_checkpoint(checkpoint_path)
            mbridge_ckpt = False
        except AssertionError:
            raise RuntimeError(f"Checkpoint at {checkpoint_path} is not in a supported format.")

    if mbridge_ckpt:
        model_cfg = instantiate(run_config["model"])
    else:
        model_cfg = _transformer_config_from_args(mlm_args)

    return model_cfg, mlm_args


def build_and_load_model(
    checkpoint_path: str,
    model_cfg: TransformerConfig,
    model_type: Optional[Literal["gpt", "mamba"]] = None,
    megatron_args: Optional[argparse.Namespace] = None,
    return_state_dict: bool = False,
    use_cpu_init: bool = False,
    skip_temp_dist_context: Optional[bool] = None,
) -> Union[Any, dict[str, torch.Tensor]]:
    """Load a Megatron model from a distributed checkpoint.

    Creates model instances and optionally a minimal distributed environment
    to load the model weights from `checkpoint_path` into the model.
    Automatically selects the appropriate distributed backend (Gloo for CPU, NCCL for GPU).

    Args:
        checkpoint_path: path to an MCore distributed checkpoint directory. Used only for
                    model weights (e.g., /path/to/model/checkpoints/iter_0000001).
        model_cfg: Model config from load_model_config(). Either a TransformerConfig or
            a model provider (e.g. GPTModelProvider) depending on source of checkpoint.
        model_type: If the checkpoint is from MegatronLM, the model type is required. Currently,
            only GPT and Mamba models are supported.
        megatron_args: If the checkpoint is from MegatronLM, this is required.
        return_state_dict: If True, return the state dict instead of model instance. Default: False.
        use_cpu_init: If True, use CPU initialization context for the model and Gloo backend.
                     If False, use GPU initialization and NCCL backend. Default: False.
        skip_temp_dist_context: If True, skip temporary distributed context setup.
                               If None, automatically skip if distributed is already initialized.
                               Default: None.

    Returns:
        The model instance with loaded weights if return_state_dict is False,
        otherwise returns a dictionary containing the full, unsharded model state_dict.
    """
    from megatron.bridge.training.checkpointing import (
        _load_model_weights_from_checkpoint,
    )
    from megatron.bridge.training.mlm_compat.arguments import _tokenizer_config_from_args
    from megatron.bridge.training.mlm_compat.model import _get_model, _gpt_provider, _mamba_provider
    from megatron.bridge.training.post_training.checkpointing import has_modelopt_state

    if has_modelopt_state(checkpoint_path):
        if hasattr(model_cfg, "restore_modelopt_state"):
            model_cfg.restore_modelopt_state = True

    def _call_model_provider(model_cfg):
        """Handles provider call for both MBridge and MLM providers."""
        if isinstance(model_cfg, ModelProviderMixin):
            if hasattr(model_cfg, "finalize"):
                model_cfg.finalize()
            return model_cfg.provide_distributed_model(wrap_with_ddp=False, use_cpu_initialization=use_cpu_init)
        else:
            assert model_type in ("gpt", "mamba"), f"model type {model_type} not supported."
            assert megatron_args is not None, "megatron_args must be provided if the checkpoint is from MegatronLM."

            # Generate the unpadded vocab size based on the tokenizer from the checkpoint
            cfg = _tokenizer_config_from_args(megatron_args)
            tokenizer = build_tokenizer(cfg)
            vocab_size = tokenizer.vocab_size

            # Re-calculate the padded vocab size based on the model config instead of the args from the checkpoint
            # to accommodate TP overrides
            megatron_args.padded_vocab_size = calculate_padded_vocab_size(
                vocab_size,
                megatron_args.make_vocab_size_divisible_by,
                model_cfg.tensor_model_parallel_size,
            )
            provider = _gpt_provider if model_type == "gpt" else _mamba_provider
            return _get_model(megatron_args, provider, model_cfg)

    # Auto-detect if we should skip temp dist context
    if skip_temp_dist_context is None:
        skip_temp_dist_context = dist.is_available() and dist.is_initialized()

    def _load_checkpoint():
        target_dtype = torch_dtype_from_mcore_config(model_cfg)
        if model_cfg.params_dtype != target_dtype:
            logger.info(f"Converting params_dtype from {model_cfg.params_dtype} to {target_dtype}")
            model_cfg.params_dtype = target_dtype

        if use_cpu_init:
            with megatron_cpu_init_context(model_cfg):
                model = _call_model_provider(model_cfg)
        else:
            model = _call_model_provider(model_cfg)

        if getattr(model_cfg, "restore_modelopt_state", False):
            from megatron.bridge.training.post_training.checkpointing import (
                load_modelopt_state,
            )

            load_modelopt_state(model, checkpoint_path)

        maybe_state_dict = _load_model_weights_from_checkpoint(
            checkpoint_path, model, return_state_dict=return_state_dict
        )

        if return_state_dict:
            del model
            return maybe_state_dict
        else:
            return model

    if skip_temp_dist_context:
        return _load_checkpoint()
    else:
        # Use appropriate backend based on initialization type
        backend = "gloo" if use_cpu_init else "nccl"
        with temporary_distributed_context(backend=backend):
            return _load_checkpoint()


def load_megatron_model(
    checkpoint_path: str,
    model_type: Optional[Literal["gpt", "mamba"]] = None,
    return_state_dict: bool = False,
    use_cpu_init: bool = False,
    skip_temp_dist_context: Optional[bool] = None,
    mp_overrides: Optional[ModelParallelKwargs] = None,
) -> Union[Any, dict[str, torch.Tensor]]:
    """Load a Megatron model from a distributed checkpoint.

    Wrapper around load_model_config() and build_and_load_model() for convenience.

    Args:
        checkpoint_path: path to an MCore distributed checkpoint directory
                          (e.g., /path/to/model/checkpoints/iter_0000001).
        model_type: If the checkpoint is from MegatronLM, the model type is required. Currently,
            only GPT and Mamba models are supported.
        return_state_dict: If True, return the state dict instead of model instance. Default: False.
        use_cpu_init: If True, use CPU initialization context for the model and Gloo backend.
                     If False, use GPU initialization and NCCL backend. Default: False.
        skip_temp_dist_context: If True, skip temporary distributed context setup.
                               If None, automatically skip if distributed is already initialized.
                               Default: None.
        mp_overrides: Optional model-parallel overrides to apply to the loaded config.
                      Only provided fields are overridden.

    Returns:
        The model instance with loaded weights if return_state_dict is False,
        otherwise returns a dictionary containing the full, unsharded model state_dict.
    """
    model_cfg, mlm_args = load_model_config(checkpoint_path)
    # If in single GPU environment, reset additional parallel settings
    model_cfg.tensor_model_parallel_size = 1
    model_cfg.pipeline_model_parallel_size = 1
    model_cfg.num_layers_in_first_pipeline_stage = None
    model_cfg.num_layers_in_last_pipeline_stage = None
    model_cfg.context_parallel_size = 1
    model_cfg.expert_model_parallel_size = 1
    model_cfg.expert_tensor_parallel_size = 1
    model_cfg.moe_extended_tp = False
    model_cfg.sequence_parallel = False
    model_cfg.perform_initialization = False
    model_cfg.virtual_pipeline_model_parallel_size = None
    model_cfg.hierarchical_context_parallel_sizes = None
    if use_cpu_init:
        model_cfg.fp8 = None
        model_cfg.fp8_param = False

    # Apply model-parallel overrides if provided
    if mp_overrides:
        for key, value in mp_overrides.items():
            if hasattr(model_cfg, key) and value is not None:
                setattr(model_cfg, key, value)

    return build_and_load_model(
        checkpoint_path, model_cfg, model_type, mlm_args, return_state_dict, use_cpu_init, skip_temp_dist_context
    )


def save_megatron_model(
    model: list[MegatronModule],
    path: Union[str, Path],
    ckpt_format: str = "torch_dist",
    hf_tokenizer_path: Optional[Union[str, Path]] = None,
    low_memory_save: bool = False,
    hf_tokenizer_kwargs: Optional[dict] = None,
) -> None:
    """Save a Megatron model in native Megatron checkpoint format without optimizer state.

    This method saves the model in Megatron's native checkpoint format, which
    can be loaded directly by Megatron for training or inference. The checkpoint
    includes the model configuration and weights, NO optimizer state or other
    artifacts.

    Args:
        model: Megatron model instance or list of instances.
        path: Directory path where the checkpoint will be saved.
        ckpt_format: Checkpoint format to use ("torch_dist" or other supported formats).
        hf_tokenizer_path: Optional HuggingFace model ID or path for tokenizer metadata.
            If provided, the tokenizer metadata will be included in the checkpoint.
        low_memory_save: If True, uses a memory-optimized save flow that:
            1. Builds the sharded state dict
            2. Expands ShardedTensorFactory objects early
            3. Clears factory data references
            4. Deletes the model from memory
            5. Forces garbage collection
            6. Saves the pre-processed state dict
            This reduces peak memory by ~50% for models with merged weights
            (e.g., gate+up projections) at the cost of destroying the model.
            Default is False, preserving the model for further use.
        hf_tokenizer_kwargs: Optional dictionary of kwargs to pass to the HuggingFace tokenizer.
            Common options include trust_remote_code=True for models with custom tokenizers.

    Example:
        >>> # Save model checkpoint
        >>> save_megatron_model(megatron_model, "./megatron_checkpoint")

        >>> # Save model checkpoint with tokenizer metadata
        >>> save_megatron_model(
        ...     megatron_model,
        ...     "./megatron_checkpoint",
        ...     hf_tokenizer_path="meta-llama/Meta-Llama-3-8B"
        ... )

        >>> # Save model checkpoint with low-memory mode (destroys model after save)
        >>> save_megatron_model(
        ...     megatron_model,
        ...     "./megatron_checkpoint",
        ...     low_memory_save=True
        ... )

    Note:
        - This method is collective and must be called by all ranks
        - The saved checkpoint can be loaded with Megatron's checkpoint loading utilities
        - The checkpoint format follows Megatron's standard structure for compatibility
        - When low_memory_save=True, the model is deleted and cannot be used afterward
    """
    # Create tokenizer config if tokenizer path is provided
    tokenizer_config = None
    if hf_tokenizer_path is not None:
        from megatron.bridge.training.tokenizers.config import TokenizerConfig

        tokenizer_config = TokenizerConfig(
            tokenizer_type="HuggingFaceTokenizer",
            tokenizer_model=str(hf_tokenizer_path),
            hf_tokenizer_kwargs=hf_tokenizer_kwargs or {},
        )

    # Get model config from the first model instance
    model_config = get_model_config(model[0])

    # Validate that the model config is a model provider
    if not isinstance(model_config, ModelProviderMixin):
        raise TypeError(
            f"Expected model config to be an instance of ModelProviderMixin, "
            f"but got {type(model_config).__name__}. "
            f"Model configs must inherit from ModelProviderMixin to ensure proper "
            f"model instantiation and configuration handling."
        )

    # Create global state for checkpointing
    state = GlobalState()
    state.cfg = ConfigContainer(
        model=model_config,
        train=None,
        optimizer=OptimizerConfig(use_distributed_optimizer=False),
        ddp=None,
        scheduler=None,
        dataset=None,
        logger=LoggerConfig(),
        tokenizer=tokenizer_config,
        checkpoint=CheckpointConfig(
            async_save=False,
            save=str(path),
            save_optim=False,
            save_rng=False,
            ckpt_format=ckpt_format,
            dist_ckpt_optim_fully_reshardable=True,
        ),
        dist=None,
    )

    if low_memory_save:
        # Low-memory save flow: process factories incrementally, freeing memory as we go
        import gc
        from dataclasses import replace

        from megatron.core.dist_checkpointing.dict_utils import (
            nested_values,
        )
        from megatron.core.dist_checkpointing.mapping import (
            ShardedTensor,
            ShardedTensorFactory,
        )
        from tqdm import tqdm

        from megatron.bridge.training.checkpointing import (
            _build_sharded_state_dict_metadata,
            generate_state_dict,
            get_rng_state,
        )
        from megatron.bridge.training.utils.pg_utils import get_pg_collection

        logger.info("[LOW_MEMORY_SAVE] Generating state dict...")

        # Get RNG state (minimal, since save_rng=False)
        pg_collection = get_pg_collection(model)
        rng_state = get_rng_state(
            data_parallel_random_init=False, ckpt_format=ckpt_format, pg_collection=pg_collection
        )

        # Build sharded state dict metadata
        sharded_sd_metadata = _build_sharded_state_dict_metadata(False, state.cfg.checkpoint)

        # Generate state dict with ShardedTensorFactory objects
        state_dict = generate_state_dict(
            state.cfg.checkpoint,
            model,
            optimizer=None,
            opt_param_scheduler=None,
            rng_state=rng_state,
            iteration=0,
            optim_sd_kwargs=dict(metadata=sharded_sd_metadata),
            model_sd_kwargs=dict(metadata=sharded_sd_metadata),
            rerun_state=None,
        )

        # Build a map from storage data_ptr to model parameter
        # This allows us to clear model params as we process factories
        storage_to_param = {}
        for m in model:
            for name, param in m.named_parameters():
                storage_ptr = param.data.untyped_storage().data_ptr()
                storage_to_param[storage_ptr] = param

        logger.info(f"[LOW_MEMORY_SAVE] Mapped {len(storage_to_param)} model parameters")

        def _clear_source_param(tensor):
            """Clear the model parameter that shares storage with this tensor."""
            storage_ptr = tensor.untyped_storage().data_ptr()
            if storage_ptr in storage_to_param:
                param = storage_to_param[storage_ptr]
                # Clear the parameter data
                param.data = torch.empty(0, device=param.device, dtype=param.dtype)
                del storage_to_param[storage_ptr]
                return True
            return False

        def _apply_factory_and_free(container, key, factory):
            """Apply a single factory, clone results, and free the source data."""
            # Get the source tensor before applying
            source_tensor = factory.data

            # Build the factory result (this creates clones for gated MLP)
            result = factory.build_fn(factory.key, factory.data, factory.replica_id, factory.flattened_range)

            # Clone all ShardedTensor data in the result to decouple from source
            if isinstance(result, dict):
                for sh_ten in nested_values(result):
                    if isinstance(sh_ten, ShardedTensor) and sh_ten.data is not None:
                        sh_ten.data = sh_ten.data.clone()
            elif isinstance(result, ShardedTensor) and result.data is not None:
                result = replace(result, data=result.data.clone())

            # Replace factory with result in container
            container[key] = result

            # NOW clear the source parameter from the model
            # This is safe because we've cloned all the data we need
            _clear_source_param(source_tensor)

        def _collect_factories(d):
            """Collect all (container, key, factory) tuples from nested dict/list."""
            factories = []
            if isinstance(d, dict):
                for k, v in list(d.items()):
                    if isinstance(v, ShardedTensorFactory):
                        factories.append((d, k, v))
                    elif isinstance(v, (dict, list)):
                        factories.extend(_collect_factories(v))
            elif isinstance(d, list):
                for i, v in enumerate(d):
                    if isinstance(v, ShardedTensorFactory):
                        factories.append((d, i, v))
                    elif isinstance(v, (dict, list)):
                        factories.extend(_collect_factories(v))
            return factories

        # Collect all factories first to get total count for progress bar
        factories = _collect_factories(state_dict)

        # Process factories with progress bar
        for idx, (container, key, factory) in enumerate(
            tqdm(factories, desc="[LOW_MEMORY_SAVE] Processing factories", unit="factory")
        ):
            _apply_factory_and_free(container, key, factory)
            # Periodic garbage collection
            if (idx + 1) % 20 == 0:
                gc.collect()
                if torch.cuda.is_available():
                    torch.cuda.empty_cache()

        # Force GC after processing all factories
        gc.collect()
        if torch.cuda.is_available():
            torch.cuda.empty_cache()

        # Collect remaining ShardedTensors for progress bar
        remaining_tensors = [
            sh_ten
            for sh_ten in nested_values(state_dict)
            if isinstance(sh_ten, ShardedTensor) and sh_ten.data is not None
        ]

        # Clone remaining ShardedTensor data with progress bar
        for idx, sh_ten in enumerate(tqdm(remaining_tensors, desc="[LOW_MEMORY_SAVE] Cloning tensors", unit="tensor")):
            source_tensor = sh_ten.data
            sh_ten.data = sh_ten.data.clone()
            _clear_source_param(source_tensor)
            if (idx + 1) % 50 == 0:
                gc.collect()

        # Clear any remaining model params not processed above
        remaining_params = len(storage_to_param)
        for m in model:
            for param in m.parameters():
                if param.data.numel() > 0:  # Not already cleared
                    param.data = torch.empty(0, device=param.device, dtype=param.dtype)
        model.clear()
        gc.collect()
        if torch.cuda.is_available():
            torch.cuda.empty_cache()

        logger.info(f"[LOW_MEMORY_SAVE] Cleared {remaining_params} remaining params, memory freed")

        # Save with pre-built state dict
        save_checkpoint(
            state=state,
            model=[],  # Empty model list since we've freed it
            optimizer=None,
            opt_param_scheduler=None,
            num_floating_point_operations_so_far=0,
            prebuilt_state_dict=state_dict,
            pg_collection=pg_collection,
        )
    else:
        # Save the checkpoint
        save_checkpoint(
            state=state,
            model=model,
            optimizer=None,
            opt_param_scheduler=None,
            num_floating_point_operations_so_far=0,
        )

    # Save tokenizer files separately if tokenizer config is provided
    if tokenizer_config is not None:
        from megatron.bridge.training.checkpointing import (
            get_checkpoint_name,
            save_tokenizer_assets,
        )

        # Build the tokenizer
        tokenizer = build_tokenizer(tokenizer_config)

        # Get the checkpoint name for step 0
        checkpoint_name = get_checkpoint_name(str(path), 0, release=False)

        # Save tokenizer files
        save_tokenizer_assets(tokenizer, tokenizer_config, checkpoint_name)


def dtype_from_str(dtype: str) -> torch.dtype:
    """Convert a string representation of a dtype to a torch.dtype.

    Handles common variations like 'fp16', 'bf16-mixed'. Defaults to float32
    for unrecognized strings.

    Args:
        dtype: The string representation (e.g., "bf16", "fp16", "float32").

    Returns:
        The corresponding torch.dtype.
    """
    if not isinstance(dtype, str):
        raise TypeError(f"Expected str, got {type(dtype)}")

    if dtype in ("float16", "fp16", "16", "16-mixed"):
        return torch.float16
    elif dtype in ("bfloat16", "bf16-mixed"):
        return torch.bfloat16
    else:
        return torch.float32


def dtype_from_hf(config: Any) -> torch.dtype:
    """Extract the torch.dtype from a Hugging Face PretrainedConfig object.

    Args:
        config: A Hugging Face model config object (must have `torch_dtype` attribute).

    Returns:
        The corresponding torch.dtype.

    Raises:
        ValueError: If the `torch_dtype` attribute is not a recognized string or torch.dtype.
        AttributeError: If the config object does not have a `torch_dtype` attribute.
    """
    if not hasattr(config, "torch_dtype"):
        raise AttributeError("Expected config to have attr `torch_dtype`")

    torch_dtype = config.torch_dtype

    if isinstance(torch_dtype, torch.dtype):
        return torch_dtype
    elif isinstance(torch_dtype, str):
        return dtype_from_str(torch_dtype)
    else:
        raise ValueError(f"torch_dtype is not of type str/torch.dtype, got {type(torch_dtype)}")



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Iterator utilities for handling virtual pipeline parallelism."""

import queue
from typing import Iterator, TypeVar, Union


DataT = TypeVar("DataT")


def make_data_iterator_list(
    model: list, data_iterator: Iterator[DataT]
) -> Union[Iterator[DataT], list[Iterator[DataT]]]:
    """Convert data iterator into form expected by Megatron with virtual pipeline parallelism.

    With interleaved/virtual pipeline parallelism, Megatron expects a list of one data
    iterator per model chunk. Each model chunk independently gets data from its data
    iterator, so we need to interact with the data iterator multiple times for each
    microbatch step. Instead of incorporating this logic into the data loader, we cache
    the iterator's output to the first model chunk and reuse it in the other model chunks.

    Args:
        model: List of model chunks (when virtual PP is used) or single-element list
        data_iterator: Iterator yielding microbatch data

    Returns:
        If model has only 1 chunk: returns the iterator as-is
        If model has multiple chunks: returns a list of iterators with caching behavior
            - First iterator in list consumes from data_iterator and caches values
            - Remaining iterators are proxies that read from the cache

    Example:
        >>> # With virtual PP size = 2 (model has 2 chunks)
        >>> iters = make_data_iterator_list(model=[chunk1, chunk2], data_iterator=iter(microbatches))
        >>> # len(iters) == 2
        >>> # Both iters[0] and iters[1] will yield the same microbatch data
        >>> batch_from_chunk0 = next(iters[0])  # Fetches from data_iterator, caches
        >>> batch_from_chunk1 = next(iters[1])  # Reads from cache, same data
    """
    # Single model chunk - no caching needed
    if not isinstance(model, list) or len(model) <= 1:
        return data_iterator

    class CachingIterator:
        """Iterator wrapper that caches values for proxy iterators.

        When the main iterator is advanced, it caches the value and distributes
        it to all registered proxy iterators.
        """

        class Proxy:
            """Proxy iterator that reads from the cache.

            Assumed to never advance past the caching iterator.
            """

            def __init__(self):
                self.cache = queue.Queue()

            def __iter__(self):
                return self

            def __next__(self):
                return self.cache.get_nowait()

        def __init__(self, iterator: Iterator[DataT]):
            self.iterator = iterator
            self.proxies = []

        def make_proxy(self):
            """Create a new proxy iterator that reads from this cache."""
            self.proxies.append(CachingIterator.Proxy())
            return self.proxies[-1]

        def __iter__(self):
            return self

        def __next__(self):
            """Advance the main iterator and cache the value for all proxies."""
            val = next(self.iterator)
            for proxy in self.proxies:
                proxy.cache.put(val)
            return val

    # Create list of iterator wrappers - one per model chunk
    # First iterator is the main caching iterator
    # Remaining iterators are proxies that read from the cache
    iters = [CachingIterator(data_iterator)]
    while len(iters) < len(model):
        iters.append(iters[0].make_proxy())

    return iters



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tokenizer-related helpers and common special-token utilities.
"""

import torch


# Common special tokens across VLM models
QWEN_TOKENS = [
    "<|im_start|>",
    "<|im_end|>",
    "<|vision_start|>",
    "<|vision_end|>",
    "<|vision_pad|>",
    "<|image_pad|>",
    "<|video_pad|>",
]
LLAVA_TOKENS = ["<image>", "<pad>"]
LLAMA_TOKENS = [
    "<|begin_of_text|>",
    "<|end_of_text|>",
    "<|finetune_right_pad_id|>",
    "<|step_id|>",
    "<|start_header_id|>",
    "<|end_header_id|>",
    "<|eom_id|>",
    "<|eot_id|>",
    "<|python_tag|>",
    "<|image|>",
]
GEMMA_TOKENS = ["<image_soft_token>"]

GEMMA_3N_TOKENS = [
    "<image_soft_token>",
    "<audio_soft_token>",
    "<start_of_audio>",
    "<start_of_image>",
    "<end_of_audio>",
    "<end_of_image>",
]

PAD_TOKENS = set(QWEN_TOKENS + LLAVA_TOKENS + LLAMA_TOKENS + GEMMA_TOKENS + GEMMA_3N_TOKENS)


def extract_skipped_token_ids(processor):
    """
    Returns list of tokens to mask in labels.

    Extracted from NeMo's HFAutoModelForImageTextToText.extract_skipped_token_ids
    """
    if processor is None:
        return torch.IntTensor([])
    tokenizer = getattr(processor, "tokenizer", processor)

    skipped_token_ids = []
    for key, val in tokenizer.added_tokens_decoder.items():
        if str(val) in PAD_TOKENS:
            skipped_token_ids.append(key)

    return torch.IntTensor(list(set(skipped_token_ids)))


def json2token(obj, sort_json_key: bool = True):
    """
    Convert an ordered JSON object into a token sequence.

    From NeMo's automodel_datasets.py
    """
    if type(obj) is dict:
        if len(obj) == 1 and "text_sequence" in obj:
            return obj["text_sequence"]
        output = ""
        keys = sorted(obj.keys(), reverse=True) if sort_json_key else obj.keys()
        for k in keys:
            output += rf"<s_{k}>" + json2token(obj[k], sort_json_key) + rf"</s_{k}>"
        return output
    if type(obj) is list:
        return r"<sep/>".join([json2token(item, sort_json_key) for item in obj])
    return str(obj)


def process_text_batch(
    processor,
    texts: list[str],
    images: list | None = None,
) -> dict[str, torch.Tensor]:
    """
    Process a batch of texts and optionally images.

    Args:
        processor: The processor to use for tokenization and image processing
        texts: List of text strings to process
        images: Optional list of images to process

    Returns:
        Dict containing processed batch data
    """
    if images is not None:
        batch = processor(
            text=texts,
            images=images,
            padding=True,
            return_tensors="pt",
        )
        if "pixel_values" in batch:
            batch["pixel_values"] = batch["pixel_values"].to(torch.bfloat16)
    else:
        batch = processor(
            text=texts,
            padding=True,
            return_tensors="pt",
        )

    return batch



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
VLM dataset utilities.

Public API re-exports:
- Makers: functions to build conversation examples from HF datasets
- Providers: classes that build PyTorch datasets bound to HF processors
- Collate fns: model-specific batch builders
"""

from megatron.bridge.data.energon.energon_provider import EnergonProvider
from megatron.bridge.data.vlm_datasets.collate import (
    COLLATE_FNS,
    default_collate_fn,
    nemotron_nano_v2_vl_collate_fn,
    phi4_mm_collate_fn,
    qwen2_5_collate_fn,
)
from megatron.bridge.data.vlm_datasets.conversation_dataset import VLMConversationDataset
from megatron.bridge.data.vlm_datasets.hf_dataset_makers import (
    make_cord_v2_dataset,
    make_cv17_dataset,
    make_llava_video_178k_dataset,
    make_medpix_dataset,
    make_raven_dataset,
    make_rdr_dataset,
)
from megatron.bridge.data.vlm_datasets.hf_provider import HFDatasetConversationProvider
from megatron.bridge.data.vlm_datasets.mock_provider import MockVLMConversationProvider
from megatron.bridge.data.vlm_datasets.preloaded_provider import PreloadedVLMConversationProvider


__all__ = [
    # Makers
    "make_rdr_dataset",
    "make_cord_v2_dataset",
    "make_medpix_dataset",
    "make_cv17_dataset",
    "make_raven_dataset",
    "make_llava_video_178k_dataset",
    # Dataset types/providers
    "VLMConversationDataset",
    "HFDatasetConversationProvider",
    "PreloadedVLMConversationProvider",
    "MockVLMConversationProvider",
    "EnergonProvider",
    # Collation utilities
    "COLLATE_FNS",
    "default_collate_fn",
    "qwen2_5_collate_fn",
    "phi4_mm_collate_fn",
]



# Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.

"""Dataloaders."""

import random
from typing import Any, Callable, Iterator, Optional

import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset


def build_pretraining_data_loader(
    dataset: Dataset,
    consumed_samples: int,
    dataloader_type: str,
    micro_batch_size: int,
    num_workers: int,
    data_sharding: bool,
    worker_init_fn: Optional[Callable] = None,
    collate_fn: Optional[Callable] = None,
    pin_memory: bool = True,
    persistent_workers: bool = False,
    data_parallel_rank: int = 0,
    data_parallel_size: int = 1,
    drop_last: Optional[bool] = True,
    global_batch_size: Optional[int] = None,
) -> Optional[DataLoader]:
    """Build a dataloader for pretraining.

    Selects the appropriate sampler (MegatronPretrainingSampler,
    MegatronPretrainingRandomSampler, or MegatronPretrainingBatchSampler)
    based on `dataloader_type` and constructs a PyTorch DataLoader.

    Args:
        dataset: The dataset to load data from.
        consumed_samples: The number of samples already consumed (for resuming).
        dataloader_type: Type of dataloader: 'single', 'cyclic', 'batch', or 'external'.
                         'external' passes the dataset through directly.
        micro_batch_size: The batch size per GPU.
        num_workers: Number of workers for the DataLoader.
        data_sharding: Whether data sharding is enabled (used for random sampler).
        worker_init_fn: Optional function to initialize workers.
        collate_fn: Optional custom collate function.
        pin_memory: Whether to pin memory for the DataLoader.
        persistent_workers: Whether to use persistent workers.
        data_parallel_rank: Rank of the current GPU in the data parallel group.
        data_parallel_size: Total number of GPUs in the data parallel group.
        drop_last: Whether to drop last incomplete batch.
        global_batch_size: Total batch size across all data parallel ranks.
                          Required for 'batch' dataloader_type.

    Returns:
        A PyTorch DataLoader instance, or the dataset itself if dataloader_type is
        'external', or None if the input dataset is None.

    Raises:
        Exception: If an unsupported dataloader_type is provided.
        RuntimeError: If 'batch' dataloader_type is used without global_batch_size.
    """

    if dataset is None:
        return None

    # Megatron sampler
    if dataloader_type == "single":
        batch_sampler = MegatronPretrainingSampler(
            total_samples=len(dataset),
            consumed_samples=consumed_samples,
            micro_batch_size=micro_batch_size,
            data_parallel_rank=data_parallel_rank,
            data_parallel_size=data_parallel_size,
            drop_last=drop_last,
        )
    elif dataloader_type == "cyclic":
        batch_sampler = MegatronPretrainingRandomSampler(
            dataset,
            total_samples=len(dataset),
            consumed_samples=consumed_samples,
            micro_batch_size=micro_batch_size,
            data_parallel_rank=data_parallel_rank,
            data_parallel_size=data_parallel_size,
            data_sharding=data_sharding,
        )
    elif dataloader_type == "batch":
        if global_batch_size is None:
            raise RuntimeError(
                "global_batch_size must be provided when using dataloader_type='batch'. "
                "This is required for proper batch sampling in fine-tuning scenarios."
            )
        batch_sampler = MegatronPretrainingBatchSampler(
            total_samples=len(dataset),
            consumed_samples=consumed_samples,
            micro_batch_size=micro_batch_size,
            global_batch_size=global_batch_size,
            data_parallel_rank=data_parallel_rank,
            data_parallel_size=data_parallel_size,
            drop_last=drop_last,
            pad_samples_to_global_batch_size=not drop_last,
        )
    elif dataloader_type == "external":
        # External dataloaders are passed through. User is expected to provide a
        # torch-compatible dataloader and define samplers, if needed.
        return dataset
    else:
        raise Exception("{} dataloader type is not supported.".format(dataloader_type))

    # Torch dataloader.
    return DataLoader(
        dataset,
        batch_sampler=batch_sampler,
        num_workers=num_workers,
        pin_memory=pin_memory,
        collate_fn=collate_fn,
        persistent_workers=persistent_workers,
        worker_init_fn=worker_init_fn,
    )


class MegatronPretrainingSampler:
    """Batch sampler for Megatron pretraining (sequential, non-random).

    Provides indices for microbatches for a specific data parallel rank,
    ensuring that data is processed sequentially across ranks and iterations.

    Args:
        total_samples: Total number of samples in the dataset.
        consumed_samples: Number of samples already consumed (for resuming).
        micro_batch_size: Batch size per GPU.
        data_parallel_rank: Rank of the current GPU in the data parallel group.
        data_parallel_size: Total number of GPUs in the data parallel group.
        drop_last (bool, optional): If True, drops the last incomplete batch.
                                  Defaults to True.
    """

    def __init__(
        self,
        total_samples: int,
        consumed_samples: int,
        micro_batch_size: int,
        data_parallel_rank: int,
        data_parallel_size: int,
        drop_last: bool = True,
    ) -> None:
        # Keep a copy of input params for later use.
        self.total_samples = total_samples
        self.consumed_samples = consumed_samples
        self.micro_batch_size = micro_batch_size
        self.data_parallel_rank = data_parallel_rank
        self.micro_batch_times_data_parallel_size = self.micro_batch_size * data_parallel_size
        self.drop_last = drop_last

        # Sanity checks.
        assert self.total_samples > 0, "no sample to consume: {}".format(self.total_samples)
        assert self.consumed_samples < self.total_samples, "no samples left to consume: {}, {}".format(
            self.consumed_samples, self.total_samples
        )
        assert self.micro_batch_size > 0
        assert data_parallel_size > 0
        assert self.data_parallel_rank < data_parallel_size, (
            "data_parallel_rank should be smaller than data size: {}, {}".format(
                self.data_parallel_rank, data_parallel_size
            )
        )

    def __len__(self) -> int:
        """Return the total number of samples."""
        return self.total_samples

    def get_start_end_idx(self) -> tuple[int, int]:
        """Calculate the start and end index for the current rank's microbatch."""
        start_idx = self.data_parallel_rank * self.micro_batch_size
        end_idx = start_idx + self.micro_batch_size
        return start_idx, end_idx

    def __iter__(self) -> Iterator[list[int]]:
        """Yields lists of indices for each microbatch assigned to this rank."""
        batch = []
        # Last batch will be dropped if drop_last is not set False
        for idx in range(self.consumed_samples, self.total_samples):
            batch.append(idx)
            if len(batch) == self.micro_batch_times_data_parallel_size:
                start_idx, end_idx = self.get_start_end_idx()
                yield batch[start_idx:end_idx]
                batch = []

        # Check the last partial batch and see drop_last is set
        if len(batch) > 0 and not self.drop_last:
            start_idx, end_idx = self.get_start_end_idx()
            yield batch[start_idx:end_idx]


class MegatronPretrainingBatchSampler:
    """Batch sampler for Megatron fine-tuning (global batch sampling).

    Samples entire global batches before distributing indices to data parallel ranks
    in an interleaved fashion. This ensures sequences within a single global batch
    are padded to the same length, which is critical for fine-tuning with variable
    sequence lengths.

    Args:
        total_samples: Total number of samples in the dataset.
        consumed_samples: Number of samples already consumed (for resuming).
        micro_batch_size: Batch size per GPU.
        global_batch_size: Total batch size across all data parallel ranks.
        data_parallel_rank: Rank of the current GPU in the data parallel group.
        data_parallel_size: Total number of GPUs in the data parallel group.
        drop_last: If True, drops the last incomplete batch.
        pad_samples_to_global_batch_size: If True, pads incomplete batches with -1 indices.
    """

    def __init__(
        self,
        total_samples: int,
        consumed_samples: int,
        micro_batch_size: int,
        global_batch_size: int,
        data_parallel_rank: int,
        data_parallel_size: int,
        drop_last: bool = True,
        pad_samples_to_global_batch_size: bool = False,
    ) -> None:
        self.total_samples = total_samples
        self.consumed_samples = consumed_samples
        self.micro_batch_size = micro_batch_size
        self.data_parallel_rank = data_parallel_rank
        self.data_parallel_size = data_parallel_size
        self.drop_last = drop_last
        self.pad_samples_to_global_batch_size = pad_samples_to_global_batch_size
        self.micro_batch_times_data_parallel_size = self.micro_batch_size * data_parallel_size

        assert self.total_samples > 0, "no sample to consume: {}".format(self.total_samples)
        assert self.micro_batch_size > 0, f"micro_batch_size must be greater than 0, but {self.micro_batch_size}"
        assert data_parallel_size > 0, f"data parallel size must be greater than 0, but {data_parallel_size}"
        assert self.data_parallel_rank < data_parallel_size, (
            "data_parallel_rank should be smaller than data size: {}, {}".format(
                self.data_parallel_rank, data_parallel_size
            )
        )

        # Set and validate global batch size
        self._global_batch_size = global_batch_size
        if self._global_batch_size % self.micro_batch_times_data_parallel_size != 0:
            raise RuntimeError(
                f"`global_batch_size` ({self._global_batch_size}) is not divisible by "
                f"`micro_batch_size ({self.micro_batch_size}) x data_parallel_size "
                f"({self.data_parallel_size})`"
            )
        self._num_micro_batches = self._global_batch_size // self.micro_batch_times_data_parallel_size
        self._global_batch_size_on_this_data_parallel_rank = self._num_micro_batches * self.micro_batch_size

    def __len__(self) -> int:
        """Return the number of batches this sampler will yield.

        Since we now yield the full global batch at once (not split into microbatches),
        this returns the number of global batches.
        """
        num_available_samples = self.total_samples - self.consumed_samples % self.total_samples
        if self.drop_last:
            num_global_batches = num_available_samples // self._global_batch_size
        else:
            num_global_batches = (num_available_samples + self._global_batch_size - 1) // self._global_batch_size

        # Each call to __iter__ yields one global batch
        return num_global_batches

    def __iter__(self) -> Iterator[list[int]]:
        """Yields lists of indices for the full global batch assigned to this rank.

        Accumulates a full global batch, then distributes indices in interleaved fashion
        to data parallel ranks, yielding ALL indices for this rank at once. This allows
        the DataLoader's collate_fn to receive the full global batch and determine optimal
        padding across all samples before the training loop splits into microbatches.

        This is essential for variable-length finetuning where we need to:
        1. Compute max_length across the entire global batch
        2. Pad all samples to the same length
        3. Then split into microbatches with consistent sequence length
        """
        batch = []
        # Last batch will be dropped if drop_last is True
        for idx in range(self.consumed_samples % self.total_samples, self.total_samples):
            batch.append(idx)
            if len(batch) == self._global_batch_size:
                # Distribute indices in interleaved fashion across ranks
                all_indices = [
                    batch[i]
                    for i in range(
                        self.data_parallel_rank,
                        self._global_batch_size,
                        self.data_parallel_size,
                    )
                ]
                assert len(all_indices) == self._global_batch_size_on_this_data_parallel_rank

                # Yield ALL indices at once (not split into microbatches)
                # The training loop will handle splitting after collation
                yield all_indices

                batch = []

        # Check the last partial batch and see if drop_last is set
        if len(batch) > 0 and not self.drop_last:
            # Distribute partial batch in interleaved fashion
            all_indices = [batch[i] for i in range(self.data_parallel_rank, len(batch), self.data_parallel_size)]
            if self.pad_samples_to_global_batch_size:
                num_pad = self._global_batch_size // self.data_parallel_size - len(all_indices)
                all_indices = all_indices + [-1] * num_pad

            # Yield ALL indices at once
            yield all_indices


class RandomSeedDataset(Dataset):
    """A dataset wrapper that sets the random seed based on epoch and index.

    Ensures reproducibility for random operations within the dataset's __getitem__
    when using multiple workers.

    Args:
        dataset: The base dataset to wrap.
        seed: The base random seed.
    """

    def __init__(self, dataset: Dataset, seed: int) -> None:
        """Initialize RandomSeedDataset."""
        self.base_seed = seed
        self.curr_seed = seed
        self.dataset = dataset

    def __len__(self) -> int:
        """Return the length of the base dataset."""
        return len(self.dataset)

    def set_epoch(self, epoch: int) -> None:
        """Set the current epoch number to adjust the random seed."""
        self.curr_seed = self.base_seed + epoch

    def __getitem__(self, idx: int) -> Any:
        """Get an item from the dataset, setting the random seed first."""
        seed = idx + self.curr_seed
        torch.manual_seed(seed)
        random.seed(seed)
        np.random.seed(seed)
        return self.dataset[idx]


class MegatronPretrainingRandomSampler:
    """Batch sampler for Megatron pretraining (randomized).

    Provides indices for microbatches for a specific data parallel rank,
    randomizing the order of samples within each epoch while supporting resumption.
    Handles data sharding across ranks if enabled.

    Args:
        dataset: The dataset (potentially wrapped with RandomSeedDataset).
        total_samples: Total number of samples in the dataset.
        consumed_samples: Number of samples already consumed (for resuming).
        micro_batch_size: Batch size per GPU.
        data_parallel_rank: Rank of the current GPU in the data parallel group.
        data_parallel_size: Total number of GPUs in the data parallel group.
        data_sharding: Whether data sharding is enabled.
    """

    def __init__(
        self,
        dataset: Dataset,
        total_samples: int,
        consumed_samples: int,
        micro_batch_size: int,
        data_parallel_rank: int,
        data_parallel_size: int,
        data_sharding: bool,
    ) -> None:
        # Keep a copy of input params for later use.
        self.dataset = dataset
        self.total_samples = total_samples
        self.consumed_samples = consumed_samples
        self.micro_batch_size = micro_batch_size
        self.data_parallel_rank = data_parallel_rank
        self.data_parallel_size = data_parallel_size
        self.data_sharding = data_sharding
        self.micro_batch_times_data_parallel_size = self.micro_batch_size * data_parallel_size
        self.last_batch_size = self.total_samples % self.micro_batch_times_data_parallel_size

        # Sanity checks.
        assert self.total_samples > 0, "no sample to consume: {}".format(self.total_samples)
        assert self.micro_batch_size > 0
        assert data_parallel_size > 0
        assert self.data_parallel_rank < data_parallel_size, (
            "data_parallel_rank should be smaller than data size: {}, {}".format(
                self.data_parallel_rank, data_parallel_size
            )
        )

    def __len__(self) -> int:
        """Return the total number of samples."""
        return self.total_samples

    def __iter__(self) -> Iterator[list[int]]:
        """Yields lists of indices for each microbatch assigned to this rank.

        Handles randomization within an epoch and data sharding.
        """
        active_total_samples = self.total_samples - self.last_batch_size
        self.epoch = self.consumed_samples // active_total_samples
        current_epoch_samples = self.consumed_samples % active_total_samples
        assert current_epoch_samples % self.micro_batch_times_data_parallel_size == 0

        if isinstance(self.dataset, RandomSeedDataset):
            self.dataset.set_epoch(self.epoch)

        # data sharding and random sampling
        if self.data_sharding:
            bucket_size = (self.total_samples // self.micro_batch_times_data_parallel_size) * self.micro_batch_size
            bucket_offset = current_epoch_samples // self.data_parallel_size
            start_idx = self.data_parallel_rank * bucket_size

            g = torch.Generator()
            g.manual_seed(self.epoch)
            random_idx = torch.randperm(bucket_size, generator=g).tolist()
            idx_range = [start_idx + x for x in random_idx[bucket_offset:]]
        else:
            full_bucket_size = (self.total_samples // self.micro_batch_size) * self.micro_batch_size
            full_bucket_offset = current_epoch_samples
            g = torch.Generator()
            g.manual_seed(self.epoch)
            idx_range_total = torch.randperm(full_bucket_size, generator=g).tolist()
            idx_range_active = idx_range_total[full_bucket_offset:]
            idx_range = idx_range_active[self.data_parallel_rank :: self.data_parallel_size]

        batch = []
        # Last batch if not complete will be dropped.
        for idx in idx_range:
            batch.append(idx)
            if len(batch) == self.micro_batch_size:
                self.consumed_samples += self.micro_batch_times_data_parallel_size
                yield batch
                batch = []
